import numpy as np
from radtran.atmosphere import Atmosphere
from sasktranif import sasktranif as skif


class AtmosphereTwoDim(Atmosphere):
    """
    The same as Atmosphere, except we also hold a two dimensional grid and
    species on that grid

    It is not recommended to use this class for general purpose 2d/3d radiative
    transfer calculations.   This class is specific to the tomographic retrieval
    code

    """

    def __init__(self):
        super().__init__()
        self.retrieval_grid = dict()
        self.user_guid_twodim = dict()
        self.user_optprop_twodim = dict()
        self.user_profile_twodim = dict()  # [angle, altitude]
        self.angleidx = None
        self.twodimalbedo = None

        self.user_defined_background = dict()

    def add_user_defined_twodim(self, grid, guid, optprop, profile, ident):
        """
        Adds a user defined two dimensional profile to the atmosphere

        Parameters
        ----------
        grid : TwoDimGrid
            From retrieval.ret2d.grid, specifies the locations of the profile
        guid :
            guid from sasktranif
        optprop : ISKOpticalProperty
        profile : 2d numpy array
            Shape (angles, heights) where angles and heights are specified in grid
        ident : string
            User specified string identifier for the profile
        """
        self.retrieval_grid[ident] = grid
        self.user_guid_twodim[ident] = guid
        self.user_optprop_twodim[ident] = optprop
        self.user_profile_twodim[ident] = profile

    def set_user_defined_background(self, grid, optprop, alts, temperature, pressure, numden):
        self.user_defined_background['numden'] = numden
        self.user_defined_background['alts'] = alts
        self.user_defined_background['pressure'] = pressure
        self.user_defined_background['temperature'] = temperature
        self.user_defined_background['optprop'] = optprop
        self.user_defined_background['grid'] = grid

    def get_user_defined_twodim(self, ident):
        """
        Returns a user defined two dimensional profile

        Parameters
        ----------
        ident : string
            Identifier for the request profile

        Returns
        -------
        2d numpy array
            shape (angles, heights)

        Raises
        ------
        KeyError
            If no profile is found matching ident

        """
        return self.user_profile_twodim[ident]

    def modify_user_defined_twodim(self, ident, profile):
        """
        Modifies the user defined profile specified by ident

        Parameters
        ----------
        ident : string
            User specified identifier for the profile to be modified
        profile : 2d numpy array
            New values for the profile, shape (angles, grid)
        """
        self.user_profile_twodim[ident] = profile

    def set_twodim_albedo(self, albedo):
        self.twodimalbedo = albedo

    def add_user_defined_background(self, engine, angles, angleidx, normalandreference):
        clim = skif.ISKClimatology('USERDEFINED_PROFILE_PLANE')
        clim.SetProperty('Heights', self.user_defined_background['alts'])
        clim.SetProperty('Angles', np.rad2deg(angles))
        clim.SetProperty('normalandreference', normalandreference)

        # temperature climatology
        clim.SetProperty('DoLogInterpolation', 0)
        clim.SetPropertyUserDefined(skif.SKCLIMATOLOGY_TEMPERATURE_K,
                                    self.user_defined_background['temperature'][angleidx, :].flatten().tolist())

        # pressure climatology
        clim.SetProperty('DoLogInterpolation', 1)
        clim.SetPropertyUserDefined(skif.SKCLIMATOLOGY_PRESSURE_PA,
                                    self.user_defined_background['pressure'][angleidx, :].flatten().tolist())

        # air number density climatology
        clim.SetProperty('DoLogInterpolation', 1)
        clim.SetPropertyUserDefined(skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3,
                                    self.user_defined_background['numden'][angleidx, :].flatten().tolist())

        engine.AddSpecies(skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3, clim, self.user_defined_background['optprop'])
        engine.SetAtmosphericState(clim)

    def add_to_engine(self, engine, reftp):
        """
        Adds the atmosphere to the engine

        Parameters
        ----------
        engine : ISKEngine
        reftp : numpy array
            Shape (3,), x,y,z coordinates of the reference point
        """
        use_def_background = len(self.user_defined_background) != 0
        super().add_to_engine(engine, addbackground=(not use_def_background))

        # Right now this will only work if the retrieval_grid is the same for all
        # species, this might change in the future so we do this loop
        for grid, optprop, guid, prof in zip(self.retrieval_grid.items(), self.user_optprop_twodim.items(),
                                             self.user_guid_twodim.items(), self.user_profile_twodim.items()):
            angles, angleidx, normalandreference = grid[1].get_local_plane(reftp)
            clim = skif.ISKClimatology('USERDEFINED_PROFILE_PLANE')
            clim.SetProperty('Heights', grid[1].altitudes.tolist())
            clim.SetProperty('Angles', np.rad2deg(angles).tolist())
            clim.SetProperty('normalandreference', normalandreference.tolist())
            clim.SetPropertyUserDefined(guid[1], prof[1][angleidx, :].flatten().tolist())
            engine.AddSpecies(guid[1], clim, optprop[1])
            engine.SetProperty('opticalanglegrid', np.rad2deg(angles).tolist())
            engine.SetProperty('opticalnormalandreference', normalandreference.tolist())
            self.angleidx = angleidx

        if use_def_background and len(self.retrieval_grid) > 0:
            self.add_user_defined_background(engine, angles, angleidx, normalandreference)

        if self.twodimalbedo is not None:
            clim = skif.ISKClimatology('USERDEFINED_PROFILE_PLANE')
            clim.SetProperty('Heights', [0, 100000])
            clim.SetProperty('Angles', np.rad2deg(angles).tolist())
            clim.SetProperty('normalandreference', normalandreference.tolist())
            clim.SetPropertyUserDefined(skif.SKCLIMATOLOGY_ALBEDO,
                                        np.tile(self.twodimalbedo[self.angleidx], (2, 1)).transpose().flatten())
            engine.SetProperty('albedoclim', clim)
