import h5py
from os import listdir
import os
import logging
import re
import numpy as np
from scipy import interpolate
import copy
import julian


def tropopause_alt(orbit, folder):
    keyword = 'ANC'
    found_folder = False
    #check if folder exists
    if os.path.isdir( folder ):
        files = listdir(folder)
        found_folder = True
    #if not check if its on the python path
    else:
        user_paths = os.environ['PYTHONPATH'].split(os.pathsep)
        for i in range(len(user_paths)):
            if os.path.isdir( user_paths[i]+folder ):
                found_folder = True
                folder = user_paths[i]+folder
                files = listdir(folder)
                break

    if found_folder is False:
        logging.error('Could not find the omps ancillary folder')
        raise Exception('Could not find the omps ancillary folder')

    filename = None
    for f in files:
            m = re.search(keyword, f)
            if m is not None:
                m = re.search( 'o([0-9]*)', f )
                orbitnum = int(m.group(1))
                if orbitnum == orbit:
                    filename = folder + f
                    break

    if filename is None:
        logging.error('Could not find the omps ancillary file')
        raise Exception('Could not find the omps ancillary file')
    else:
        #load in the atmospheric data from the OMPS ancillary files
        f = h5py.File(r''+filename,'r')
        trop_alt = f['GRIDDED_DATA/TropopauseAltitude'][:, 1]*1000      # trop_alt in m
    return trop_alt


def convert_to_ppm(numdenprofile, airdensity):
    return numdenprofile / airdensity * 1e6

def convert_to_ret_alt_grid(retalt, profalt, profile):
    return interpolate.interp1d(profalt, profile, axis=0)(retalt)

def convert_pickle_to_l2(data, out_file, out_file_raw=None, save_averaging_kernel=False):
    lat = data['retrieval_geometry', 'latitude']
    lon = data['retrieval_geometry', 'longitude']
    retrieved_images = data['omps_geometry', 'image_index']
    alt = data['retrieval_geometry', 'altitudes']
    sza = data['omps_geometry', 'sza']
    mjd = data['retrieval_geometry', 'mjd']

    o3 = data['retrieved', 'o3'].transpose()
    o3_error = data['retrieved', 'ozone_numden_error'].transpose()
    aero = data['retrieved', 'aero'].transpose()
    aero_error = data['retrieved', 'aerosol_numden_error'].transpose()
    aero_ext = data['retrieved', 'aerosol_extinction'].transpose()
    aero_ext_error = data['retrieved', 'aerosol_extinction_error'].transpose()
    aero_wavel = data['retrieved', 'aerosol_wavelength']
    albedo = data['retrieved', 'albedo']

    o3_rtm = copy.copy(data['retrieved', 'o3_RTM'].transpose())
    aero_rtm = copy.copy(data['retrieved', 'aero_RTM'].transpose())
    aero_ext_rtm = copy.copy(aero_rtm)
    albedo_rtm = copy.copy(albedo)

    o3_flag = np.isnan(o3)
    try:
        trop_alt = data['retrieved', 'trop_alt'] - 2000
        cloud_alt = data['retrieved', 'cloud_alt']

        max_alt = np.maximum(trop_alt, cloud_alt)
        for idx, ta in enumerate(max_alt):
            bad = alt < ta
            o3_flag[bad, idx] = 1
    except:
        pass

    write_aero = aero
    write_aero_ext = aero_ext

    write_temp = data['retrieval_geometry', 'airtemperature'].transpose()
    write_pressure = data['retrieval_geometry', 'airpressure'].transpose()
    write_airdensity = data['retrieval_geometry', 'airnumberdensity'].transpose()

    write_temp = convert_to_ret_alt_grid(alt, data['retrieval_geometry', 'airalt'], write_temp)
    write_pressure = convert_to_ret_alt_grid(alt, data['retrieval_geometry', 'airalt'], write_pressure)
    write_airdensity = convert_to_ret_alt_grid(alt, data['retrieval_geometry', 'airalt'], write_airdensity)

    write_o3ppm = convert_to_ppm(o3, write_airdensity)
    write_o3ppm_error = convert_to_ppm(o3_error, write_airdensity)

    aero_nan_flag = np.isnan(aero)
    try:
        trop_alt = data['retrieved', 'trop_alt'] - 6000
        cloud_alt = data['retrieved', 'cloud_alt']

        max_alt = np.maximum(trop_alt, cloud_alt)
        for idx, ta in enumerate(max_alt):
            bad = alt < ta
            aero_nan_flag[bad, idx] = 1
    except:
        pass

    aero_error[np.isnan(write_aero)] = np.nan
    aero_ext_error[np.isnan(write_aero_ext)] = np.nan

    write_albedo = albedo

    write_lat = lat

    write_lon = lon

    write_sza = sza

    write_mjd = mjd

    write_alt = alt/1000

    write_cloud_alt = data['retrieved', 'cloud_alt'] / 1000

    write_trop_alt = data['retrieved', 'trop_alt'] / 1000

    f = h5py.File(out_file, "w")

    g = f.create_group("Data")
    u = f.create_group("DataUnbounded")
    h = f.create_group("Geo")
    s = f.create_group("Software")

    dset = h.create_dataset("Latitude", data=write_lat, compression='gzip')
    dset.attrs['description'] = 'Latitude grid for the retrieval grid'
    dset.attrs['dimensions'] = 'NumProfile'
    dset.attrs['units'] = 'Degrees'

    dset = h.create_dataset("Longitude", data=write_lon, compression='gzip')
    dset.attrs['description'] = 'Longitude grid for the retrieval grid'
    dset.attrs['dimensions'] = 'NumProfile'
    dset.attrs['units'] = 'Degrees'

    # h.create_dataset("SZA", data=write_sza, compression='gzip')

    dset = h.create_dataset("MJD", data=write_mjd, compression='gzip')
    dset.attrs['description'] = 'Modified Julian Date for the retrieval grid'
    dset.attrs['dimensions'] = 'NumProfile'

    t_stam = [julian.from_jd(m, fmt='mjd').isoformat().encode("ascii", "ignore") for m in write_mjd]

    dset = h.create_dataset("ISOTime", data=t_stam)
    dset.attrs['description'] = 'ISO Time Stamp for the retrieval grid'
    dset.attrs['dimensions'] = 'NumProfile'

    dset = h.create_dataset("Altitude", data=write_alt, compression='gzip')
    dset.attrs['description'] = 'Altitude for the retrieval grid'
    dset.attrs['dimensions'] = 'NumHeight'
    dset.attrs['units'] = 'km'

    t_o3 = copy.copy(o3)
    t_o3[o3_flag] = np.nan
    dset = g.create_dataset("OzoneNumDen", data=t_o3, compression='gzip')
    dset.attrs['description'] = 'Retrieved ozone number density'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/cm^3'

    t_o3 = copy.copy(o3_error)
    t_o3[o3_flag] = np.nan
    dset = g.create_dataset("OzoneNumDenError", data=t_o3, compression='gzip')
    dset.attrs['description'] = 'Error in the retrieved ozone number density'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/cm^3'

    t_o3 = copy.copy(write_o3ppm)
    t_o3[o3_flag] = np.nan
    dset = g.create_dataset("OzonePPM", data=t_o3, compression='gzip')
    dset.attrs['description'] = 'Retrieved ozone parts per million'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = 'ppm'

    t_o3 = copy.copy(write_o3ppm_error)
    t_o3[o3_flag] = np.nan
    dset = g.create_dataset("OzonePPMError", data=t_o3, compression='gzip')
    dset.attrs['description'] = 'Error in the retrieved ozone parts per million'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = 'ppm'

    dset = g.create_dataset("BackgroundTemperature", data=write_temp, compression='gzip')
    dset.attrs['description'] = 'Background climatological temperature from MERRA on the retrieval grid'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = 'K'

    dset = g.create_dataset("BackgroundPressure", data=write_pressure, compression='gzip')
    dset.attrs['description'] = 'Background climatological pressure from MERRA on the retrieval grid'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = 'Pa'

    dset = g.create_dataset("BackgroundNumDen", data=write_airdensity, compression='gzip')
    dset.attrs['description'] = 'Background climatological number density from MERRA on the retrieval grid'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/cm^3'

    t_a = copy.copy(write_aero)
    t_a[aero_nan_flag] = np.nan
    dset = g.create_dataset("AerosolNumDen", data=t_a, compression='gzip')
    dset.attrs['description'] = 'Retrieved aerosol number density'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/cm^3'

    t_a = copy.copy(write_aero_ext)
    t_a[aero_nan_flag] = np.nan
    dset = g.create_dataset("AerosolExtinction", data=t_a, compression='gzip')
    dset.attrs['description'] = 'Retrieved aerosol extinction at Data/AerosolWavelength'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/km'

    t_a = copy.copy(aero_error)
    t_a[aero_nan_flag] = np.nan
    dset = g.create_dataset("AerosolNumDenError", data=t_a, compression='gzip')
    dset.attrs['description'] = 'Error in the retrieved aerosol number density'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/cm^3'

    t_a = copy.copy(aero_ext_error)
    t_a[aero_nan_flag] = np.nan
    dset = g.create_dataset("AerosolExtinctionError", data=t_a, compression='gzip')
    dset.attrs['description'] = 'Error in the retrieved aerosol extinction at Data/AerosolWavelength'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/km'

    dset = g.create_dataset("AerosolWavelength", data=aero_wavel)
    dset.attrs['description'] = 'Wavelength the aerosol extinction is calculated at'
    dset.attrs['dimensions'] = 'Scalar'
    dset.attrs['units'] = 'nm'

    dset = g.create_dataset("Albedo", data=write_albedo, compression='gzip')
    dset.attrs['description'] = 'Retrieved albedo assuming a Lambertian surface'
    dset.attrs['units'] = 'Unitless'
    dset.attrs['dimensions'] = 'NumProfile'

    dset = g.create_dataset("CloudHeight", data=write_cloud_alt, compression='gzip')
    dset.attrs['description'] = 'Retrieved cloud height'
    dset.attrs['dimensions'] = 'NumProfile'
    dset.attrs['units'] = 'km'

    dset = g.create_dataset("TropopauseHeight", data=write_trop_alt, compression='gzip')
    dset.attrs['description'] = 'Tropopause height from OMPS-LP L1 ANC file on the retrieval grid'
    dset.attrs['dimensions'] = 'NumProfile'
    dset.attrs['units'] = 'km'

    s.create_dataset("processing_date", data=data['processing', 'date'])
    s.create_dataset("processing_location", data="Westgrid Orcinus")
    s.create_dataset("version", data=data['processing', 'version'])

    dset = u.create_dataset("OzoneNumDen", data=o3, compression='gzip')
    dset.attrs['description'] = 'Retrieved ozone number density'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/cm^3'

    dset = u.create_dataset("OzoneNumDenError", data=o3_error, compression='gzip')
    dset.attrs['description'] = 'Error in the retrieved ozone number density'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/cm^3'

    dset = u.create_dataset("OzonePPM", data=write_o3ppm, compression='gzip')
    dset.attrs['description'] = 'Retrieved ozone parts per million'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = 'ppm'

    dset = u.create_dataset("OzonePPMError", data=write_o3ppm_error, compression='gzip')
    dset.attrs['description'] = 'Error in the retrieved ozone parts per million'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = 'ppm'

    dset = u.create_dataset("AerosolNumDen", data=write_aero, compression='gzip')
    dset.attrs['description'] = 'Retrieved aerosol number density'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/cm^3'

    dset = u.create_dataset("AerosolExtinction", data=write_aero_ext, compression='gzip')
    dset.attrs['description'] = 'Retrieved aerosol extinction at Data/AerosolWavelength'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/km'

    dset = u.create_dataset("AerosolNumDenError", data=aero_error, compression='gzip')
    dset.attrs['description'] = 'Error in the retrieved aerosol number density'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/cm^3'

    dset = u.create_dataset("AerosolExtinctionError", data=aero_ext_error, compression='gzip')
    dset.attrs['description'] = 'Error in the retrieved aerosol extinction at Data/AerosolWavelength'
    dset.attrs['dimensions'] = 'NumHeight x NumProfile'
    dset.attrs['units'] = '/km'

    f.close()

    if out_file_raw is not None:
        f = h5py.File(out_file_raw, "w")

        ret_info = f.create_group("RetInfo")
        atmosphere = ret_info.create_group("Atmosphere")
        radiances = ret_info.create_group("Radiances")
        radmeas = data['omps_measurement', 'radiance']
        wavel_meas = data['omps_measurement', 'wavel_nm']

        try:
            debug = data['debug_aerosol']

            ymeas = debug[1]['ymeas'].reshape((len(data['omps_geometry', 'image_index']), -1))
            yret = debug[1]['yret'].reshape((len(data['omps_geometry', 'image_index']), -1))

            ret_info.create_dataset("Aerosol_ymeas", data=ymeas, compression='gzip')
            ret_info.create_dataset("Aerosol_yret", data=yret, compression='gzip')

            rad_ret = debug[1]['rad_ret']
            wavel_ret = debug[1]['wavel_ret']

            rad_meas = interpolate.interp1d(wavel_meas, radmeas, axis=1, bounds_error=False)(wavel_ret)

            radiances.create_dataset("Aerosol_measrad", data=rad_meas, compression='gzip')
            radiances.create_dataset("Aerosol_retrad", data=rad_ret, compression='gzip')
            radiances.create_dataset("Aerosol_wavel", data=wavel_ret, compression='gzip')

            try:
                A = debug[1]['averaging_kernel']
                d = debug[1]['chi_squared']
                if save_averaging_kernel:
                    ret_info.create_dataset("Aerosol_averaging_kernel", data=A, compression='gzip')
                ret_info.create_dataset("Aerosol_chi_squared", data=d)
            except:
                pass

        except:
            pass
        try:
            debug = data['debug_ozone']

            ymeas = debug[1]['ymeas'].reshape((len(data['omps_geometry', 'image_index']), -1))
            yret = debug[1]['yret'].reshape((len(data['omps_geometry', 'image_index']), -1))

            ret_info.create_dataset("Ozone_ymeas", data=ymeas, compression='gzip')
            ret_info.create_dataset("Ozone_yret", data=yret, compression='gzip')

            rad_ret = debug[1]['rad_ret']
            wavel_ret = debug[1]['wavel_ret']

            rad_meas = interpolate.interp1d(wavel_meas, radmeas, axis=1, bounds_error=False)(wavel_ret)

            radiances.create_dataset("Ozone_measrad", data=rad_meas, compression='gzip')
            radiances.create_dataset("Ozone_retrad", data=rad_ret, compression='gzip')
            radiances.create_dataset("Ozone_wavel", data=wavel_ret, compression='gzip')

            try:
                A = debug[1]['averaging_kernel']
                d = debug[1]['chi_squared']

                if save_averaging_kernel:
                    ret_info.create_dataset("Ozone_averaging_kernel", data=A, compression='gzip')
                ret_info.create_dataset("Ozone_chi_squared", data=d)
            except:
                pass

        except:
            pass
        try:
            debug = data['debug_albedo']

            ymeas = debug[1]['ymeas'].reshape((len(data['omps_geometry', 'image_index']), -1))
            yret = debug[1]['yret'].reshape((len(data['omps_geometry', 'image_index']), -1))

            ret_info.create_dataset("Albedo_ymeas", data=ymeas, compression='gzip')
            ret_info.create_dataset("Albedo_yret", data=yret, compression='gzip')

            rad_ret = debug[1]['rad_ret']
            wavel_ret = debug[1]['wavel_ret']

            rad_meas = interpolate.interp1d(wavel_meas, radmeas, axis=1, bounds_error=False)(wavel_ret)

            radiances.create_dataset("Albedo_measrad", data=rad_meas, compression='gzip')
            radiances.create_dataset("Albedo_retrad", data=rad_ret, compression='gzip')
            radiances.create_dataset("Albedo_wavel", data=wavel_ret, compression='gzip')
        except:
            pass

        atmosphere.create_dataset("Aerosol_rtm", data=aero_rtm, compression='gzip')
        atmosphere.create_dataset("Ozone_rtm", data=o3_rtm, compression='gzip')
        atmosphere.create_dataset("Albedo_rtm", data=albedo_rtm, compression='gzip')

        f.close()
