import numpy as np
import h5py
import os
import re
import julian
from datetime import datetime
from scipy import interpolate
from tqdm import tqdm

RSAS_CUTOFF_MJD = julian.to_jd(datetime(year=2013, day=25, month=4), fmt='mjd')
RSAS_BEFORE = 0.3
RSAS_AFTER = 0.2

NUM_INLCUDE_ALT = 58


class FileAttribute(object):
    def __init__(self, orbital_param_name, daily_param_name, attrs):
        self.orbital_param_name = orbital_param_name
        self.daily_param_name = daily_param_name
        self.data = None
        self.attrs = []

    @staticmethod
    def _postcorrect(data, mjd_start):
        return data

    def reset(self):
        self.data = None

    def load_from_orbital_file(self, hdf_file, mjd_start):
        orbit_mjd = hdf_file['Geo/MJD'][:]
        good = (orbit_mjd > mjd_start) & (orbit_mjd < (mjd_start + 1))

        data = hdf_file[self.orbital_param_name][:]
        if len(np.shape(data)) == 2:
            data = data[:NUM_INLCUDE_ALT, good]
        elif len(np.shape(data)) == 1:
            data = data[good]
        else:
            raise ValueError()

        data = self._postcorrect(data, mjd_start)
        if self.data is None:
            self.data = data
        else:
            self.data = np.hstack((self.data, data))

    def write_to_daily_file(self, hdf_file):
        if self.data is not None:
            hdf_file[self.daily_param_name] = self.data


class RSASFileAttribute(FileAttribute):
    @staticmethod
    def _postcorrect(data, mjd_start):
        if mjd_start > RSAS_CUTOFF_MJD:
            rsas_alt = RSAS_AFTER
        else:
            rsas_alt = RSAS_BEFORE

        if len(np.shape(data)) > 1:
            return interpolate.interp1d(np.linspace(0.5, 99.5, 100)[:NUM_INLCUDE_ALT], data, axis=0, bounds_error=False)(
                                        np.linspace(0.5, 99.5, 100)[:NUM_INLCUDE_ALT] + rsas_alt)
        else:
            return data + rsas_alt


class ConstantAttribute(object):
    def __init__(self, daily_param_name, attrs, value):
        self.daily_param_name = daily_param_name
        self.attrs = attrs
        self.value = value

    def reset(self):
        pass

    def load_from_orbital_file(self, hdf_file, mjd_start):
        pass

    def write_to_daily_file(self, hdf_file):
        hdf_file[self.daily_param_name] = self.value


class DateAttribute(ConstantAttribute):
    def load_from_orbital_file(self, hdf_file, mjd_start):
        # Figure this out somewhere else?
        dt = julian.from_jd(mjd_start, fmt='mjd')
        self.value = dt.year * 10000 + dt.month*100 + dt.day


class TimeAttribute(FileAttribute):
    def load_from_orbital_file(self, hdf_file, mjd_start):
        orbit_mjd = hdf_file['Geo/MJD'][:]

        data = np.array([julian.from_jd(m, fmt='mjd').timestamp() for m in orbit_mjd])
        start_date = julian.from_jd(mjd_start, fmt='mjd')
        data -= start_date.timestamp()

        # TODO: Check that this timestamp is what we want

        if self.data is None:
            self.data = data
        else:
            self.data = np.hstack((self.data, data))


class FileHandler(object):
    def __init__(self):
        self.mjd = None
        self.file_attributes = list()
        self.file_attributes.append(RSASFileAttribute('DataUnbounded/OzoneNumDen',
                                                      'DataFields/O3Value',
                                                      []))

        self.file_attributes.append(RSASFileAttribute('DataUnbounded/OzoneNumDenError',
                                                      'DataFields/O3Error',
                                                      []))

        self.file_attributes.append(FileAttribute('Data/Albedo',
                                                  'DataFields/Albedo',
                                                  []))


        self.file_attributes.append(FileAttribute('Geo/Latitude',
                                                  'GeolocationFields/Latitude',
                                                  []))

        self.file_attributes.append(FileAttribute('Geo/Longitude',
                                                  'GeolocationFields/Longitude',
                                                  []))

        self.file_attributes.append(FileAttribute('Data/BackgroundPressure',
                                                  'AncillaryData/Pressure',
                                                  []))

        self.file_attributes.append(FileAttribute('Data/TropopauseHeight',
                                                  'AncillaryData/TropopauseAltitude',
                                                  []))

        self.file_attributes.append(RSASFileAttribute('Data/CloudHeight',
                                                      'DataFields/CloudHeight',
                                                      []))

        self.file_attributes.append(ConstantAttribute('GeolocationFields/HeightScale',
                                                      [],
                                                      np.linspace(0.5, 99.5, 100)[:NUM_INLCUDE_ALT]))

        self.file_attributes.append(DateAttribute('GeolocationFields/Date',
                                                  [],
                                                  None))

        self.file_attributes.append(TimeAttribute(None,
                                                  'GeolocationFields/Time',
                                                  [],))

    def set_start_mjd(self, mjd):
        self.mjd = mjd
        for attr in self.file_attributes:
            attr.reset()

    def load_from_orbital_file(self, hdf_file, raw_hdf_file=None):
        orbit_is_good = True
        if raw_hdf_file is not None:
            chi_squared = raw_hdf_file['RetInfo/Ozone_chi_squared'].value
            if chi_squared > 2:
                orbit_is_good = False
        if orbit_is_good:
            for attr in self.file_attributes:
                attr.load_from_orbital_file(hdf_file, self.mjd)

    def write_to_daily_file(self, hdf_file):
        for attr in self.file_attributes:
            attr.write_to_daily_file(hdf_file)


def make_mjd_map(root_folder):
    """
    Creates a dictionary used to map OMPS orbits to MJD's
    Parameters
    ----------
    l2_folder: string
        Folder of the OMPS orbital output files

    Returns
    -------
    mjd_orbit: dict
        Dictionary where mjd_orbit[orbit] = (min mjd of orbit, max mjd of orbit)
    """
    l2_folder = root_folder + '/L2/'
    mjd_orbit = dict()
    for f in os.listdir(l2_folder):
        m = re.search('([0-9]*).h5', str(f))
        if m is not None and len(m.group(1)) > 0:
            try:
                orbit = int(m.group(1))
                if orbit in mjd_orbit:
                    continue
                else:
                    fi = h5py.File(l2_folder + f, "r")
                    mjd = fi['Geo/MJD'][:]
                    mjd_orbit[orbit] = (np.min(mjd), np.max(mjd))
            except:
                pass
    return mjd_orbit


def daily_file_name(mjd):
    day_start = julian.from_jd(mjd, fmt='mjd')

    return 'USASK-' + day_start.strftime("%Ym%m%d") + '.h5'


def make_daily_file(day_start_mjd, mjd_orbit, root_folder, file_handler):
    day_start = julian.from_jd(day_start_mjd, fmt='mjd')
    day_end_mjd = day_start_mjd + 1

    orbits = [o for o, m in mjd_orbit.items() if m[1] > day_start_mjd and m[0] < day_end_mjd]

    if len(orbits) > 0:
        for orbit in sorted(orbits):
            orbit_file = h5py.File(root_folder + '/L2/o' + str(orbit) + '.h5', 'r')
            raw_orbit_file = h5py.File(root_folder + '/raw/Retinfo_' + str(orbit) + '.h5', 'r')
            file_handler.load_from_orbital_file(orbit_file, raw_orbit_file)

        daily_file = h5py.File(root_folder + '/Daily3/' + daily_file_name(day_start_mjd), 'w')

        file_handler.write_to_daily_file(daily_file)


if __name__ == "__main__":
    root_folder = r'Z:/data/OMPS/OMPS_Tomography_v1.0.2/'

    mjd_orbit = make_mjd_map(root_folder)

    mjd_start = int(np.floor(np.nanmin([m[0] for o, m in mjd_orbit.items()])))
    mjd_end = int(np.ceil(np.nanmax([m[1] for o, m in mjd_orbit.items()])))

    file_handler = FileHandler()

    for mjd in tqdm(range(mjd_start, mjd_end)):
        file_handler.set_start_mjd(mjd)
        make_daily_file(mjd, mjd_orbit, root_folder, file_handler)