from os import listdir
import os
import copy as copy
import re
import logging

import numpy as np

import scipy as sp
from scipy.interpolate import InterpolatedUnivariateSpline

from tomography.omps_processing.omps_util import tropopause_alt

import sasktranif.sasktranif as skif
import tomography.atmosphere.atmo_2d
import tomography.sasktran_2d
import tomography.species.tomographicspecies as tomospecies
import tomography.species.tomographicalbedo as tomoalbedo
import retrieval.method.lsq as lsq
import retrieval.method.rodgers as rodgers
import retrieval.core as core
import radtran.options as opt
import retrieval.instruments.ideal_spectrograph as idealspectrograph
import radtran.climatologies.ozone as ozone
import radtran.climatologies.aerosol as aerosol
import radtran.climatologies.no2 as no2
import radtran.climatologies.utility as utility
import radtran.climatologies.background as background
import radtran.geometry as geometry
import omps.omps_model as omps
import omps.level2 as omps_l2
import omps.cloud_detection as cd
import time
import h5py
import utility as uutil
import xarray as xr


class OMPSTomography(object):
    """
    Class which handles the full retrieval for a single OMPS orbit

    Typically, the class is constructed with an OMPS orbit number. The user
    then changes any desired options, and finally :func:`~OMPSTomography.retrieve_orbit`
    is called, starting the retrieval.


    Parameters
    ----------
    orbitnumber : scalar
        The OMPS orbit number to retrieve from

    Examples
    --------
    A simple retrieval with default settings

    >>> omps = OMPSTomography(12112)
    >>> output = omps.retrieve_orbit()
    >>> plt.pcolor(output['retrieved','o3'].transpose())

    Retrieve only ozone with single scatter only

    >>> omps = OMPSTomography(12112)
    >>> omps.ret_opt['retrieval_order'] = ['ozone']
    >>> omps.ret_opt['max_iter'] = [5]
    >>> omps.ret_opt['scatterorder'] = 1
    >>> output = omps.retrieve_orbit()

    Options
    -------
    Options are set by accessing the ret_opt dictionary within the class,
    e.g.,

    >>> omps = OMPSTomography(12112)
    >>> omps.ret_opt['max_sza'] = 90

    All options are optional

    max_sza : scalar
        All images with a tangent point sza above this value are thrown out and not
        used in the retrieval. Default is 88

    scatterorder : scalar
        The number of orders of scatter to use in the radiative transfer model.
        Default is 30
    
    num_diffuse_profiles : scalar
        The number of diffuse profiles to use in the radiative transfer model.
        Default is 5

    retrieval_order : list of strings
        A combination of the strings 'albedo', 'aerosol', 'ozone' which determines
        the number of retrievals to run and their order.  The default is
        ['albedo', 'aerosol', 'ozone', 'albedo', 'ozone']

    max_iter : list of scalars
        The numbe of iterations to do for each of the retrievals, should be the
        same length as retrieval_order.  Default is [5, 5, 5, 2, 2]

    images : range
        The images to use in the retrieval.  Default is range(0,180).  Typically
        every OMPS orbit contains at most 180 images.

    verbose_output : bool
        If true, then intermediate results such as the measurement vectors will
        be saved in the output structure. Default is False

    maxdifference_seconds : scalar
        Successive images are modelled at the same time in the radiative transfer
        model.  However this assumes the sun is stationary. This paremeter determines
        the time period in which we can do this combination.  Default is 100 seconds

    omps_l1_folder : string
        The folder name where the OMPS L1 files are stored, default is
        r"\\utls\\users\\scratch\\OMPS_L1\\"

    omps_l1_anc_folder : string
        The folder name where the OMPS L1 anc files are stored.

    omps_l2_folder : string
        The folder name where the OMPS L2 files are stored, default is
        r"\\utls\\users\\scratch\\OMPS_L2\\"  The level 2 files are only necessary
        if the retrieval options 'use_omps_l2_as_apriori' is set to True (default is False)

    ozone_apriori_type : ('constant', 'omps_l2', 'mart', 'labow')
        Ozone apriori to use in the retrieval.  'const' uses a constant value for
        the apriori profiles.  'omps_l2' uses the reported omps_l2 values, this is
        currently buggy since extrapolation has to be done below/above where omps
        reports the values.  'mart' uses values retrieved by a 1-d retrieval.
        'labow' takes the values from the labow ozone climatology. Default value is 'labow'

    aerosol_apriori_type : ('default', 'omps_l2', 'mart')
        Aerosol apriori to use in the retrieval.  'default' uses a constant value for
        the apriori profiles.  'omps_l2' uses the reported omps_l2 values, this is
        currently buggy since extrapolation has to be done below/above where omps
        reports the values.  'mart' uses values retrieved by a 1-d retrieval.

    mart_folder : string
        Folder where the MART 1d retrieval results are stored for use as the apriori

    grid_type : ('scan', 'uniform')
        Type of retrieval grid to use, 'scan' sets a profile at the average tangent point
        for each scan or image.  'uniform' constructs a uniform grid between the
        first and last scans or images.  Default is 'uniform'

    grid_down_factor : scalar
        The ratio (num_scans) / (num_retrieval_locations).  Default is 1.  Only
        has an effect when grid_type is set to 'uniform'

    tropopause_lowerbound : bool
        If True, the tropopause altitude from the OMPS L2 file is used as the
        lower bound for retrievals.  Currently we retrieve 2 km below the tropopause for ozone and
        6 km below the tropopause for aerosol. Default is True

    cloud_lowerbound : bool
        If True, the cloud height is used to calculate the lowerbound of the retrieval.  If tropopause_lowerbound
        is also True then the maximum of the two altitudes is used as the lower bound of the retrieval.  Note, that
        even if this is set to false the cloud heights are still reported in the output file, they are just not used
        in the retrieval.
        Defaul is True.

    ozone_apriori_scaling : bool
        If True, the apriori profile is scaled the same way as the lowest altitude
        in the retrieval.  If False the apriori stays constant.  Default is True

    aerosol_smooth_factor : int
        The final retrieved aerosol profile is put through a weighted average filter
        in the image number dimension.  We do this because the aerosol tomography
        is fairly unstable (due to a lack of information compared to ozone) but
        we need a semi accurate aerosol shape to do the ozone retrieval.
        Default is 3, it must be odd. A good number here depends on the 'grid_down_factor'
        parameter, when retrieving on every image 'aerosol_smooth_factor' should
        be closer to 5, 3 seems to work well when retrieving on every second image.
        Default is '0.0'
    
    albedo_smooth_factor : int
        The albedo can optionally be resampled using a fourier transform technique.
        A gaussian window is applied in with a standard deviation of 
        (num_ret_prof) / (albedo_smooth_factor).  A value of 5 here is a fairly
        modest smoothing, while a value of 10 is a large amount of smoothing.
        Default is 10 ( large smoothing )

    model_spectrum : bool
        If True, at the end of the retrieval a full radiance calculation over all omps
        wavelengths will be done and stored in the output dictionary.  Default is False

    o3_horiz_tikh_factor : scalar
        If nonzero, a second derivative Tikhonov operator is added to the ozone retrieval.  This forcibly
        reduces oscillations in the retrieval.  Default is 2.0.

    aero_horiz_tikh_factor : scalar
        If nonzero, a second derivative Tikhonov operator is added to the aerosol retrieval.  This forcibly
        reduces oscillations in the retrieval.  Default is 4.0.

    o3_vertical_tikh_factor : scalar
        If nonzero, a second derivative Tikhonov operator is added to the ozone retrieval.  This forcibly
        reduces oscillations in the retrieval.  Default is 2.0.

    aero_vertical_tikh_factor : scalar
        If nonzero, a second derivative Tikhonov operator is added to the aerosol retrieval.  This forcibly
        reduces oscillations in the retrieval.  Default is 4.0.

    retrieval_method : string
        Either 'lm' or 'rodgers', determines the method used to solve the system

    ozone_lm_damping : int
        An LM style damping matrix is added to both the LM and rodgers retrieval.  For the ozone retrieval it is
        multiplied by this constant. Default is 0.1.  Note that the ozone damping factor is slightly higher than
        the aerosol damping factor.  It was found that the Ozone retrieval tends to be more sensitive to edge effects
        in the retrieval and this helps to minimize them.

    aerosol_lm_damping : int
        An LM style damping matrix is added to both the LM and rodgers retrieval.  For the ozone retrieval it is
        multiplied by this constant. Default is 0.01
        

    Output
    ------
    The output structure is a dictionary, typically indexed by a tuple of two strings. Here
    we list the keys and the corresponding item in the dictionary.  E.g.

    >>> output = omps.retrieve_orbit()
    >>> sza = output['omps_geometry', 'sza']

    ('omps_geometry','image_index') : 1d numpy array
        The indicies of the images used in the retrieval, this is typically something
        like 10:170, since images at the end are thrown out due to the SZA being too
        high
    ('omps_geometry', 'sza') : 1d numpy array
        Solar Zenith Angle reported from the OMPS L1 file.  This is the same size
        as ('omps_geometry', 'image_index')
    ('omps_geometry', 'longitude') : 1d numpy array
        Longitude as reported from the OMPS L1 file.  Same size as ('omps_geometry', 'image_index')
    ('omps_geometry', 'latitude') : 1d numpy array
        Latitude as reported from the OMPS L1 file.  Same size as ('omps_geometry', 'image_index')
    ('omps_geometry', 'trop_alt') : 1d numpy array
        Tropopause altitude as reported from the OMPS L1 file.  Same size as ('omps_geometry', 'image_index')
    ('omps_geometry', 'cloud_alt') : 1d numpy array
        Cloud altitude in km as calculated from the OMPS L1 file. Same size as ('omps_geometry', 'image_index')
    ('omps_measurement', 'radiance') : 2d numpy array
        Shape (los,wavel) of radiances measured by OMPS.  In the standard
        photons/cm^2/s/sterad/nm
    ('omps_measurement', 'wavel_nm') : 1d numpy array
        Wavelengths corresponding to the OMPS measurements
    ('omps_measurement', 'psf') : 1d numpy array
        Point spread function of the omps measurements assuming a gaussian shape
    ('retrieval_geometry', 'altitude') : 1d numpy array
        Altitudes in m of the retrieval grid
    ('retrieval_geometry', 'latitude') : 1d numpy array
        Latitude of the retrieval locations
    ('retrieval_geometry', 'longitude') : 1d numpy array
        Longitude of the retrieval locations
    ('retrieval_geometry', 'location_xyz') : 2d numpy array
        Array of shape(numretprof,3) of the xyz locations of the retrieval
        points
    ('retrieved', 'o3') : 2d numpy array
        Retrieved ozone.  If no ozone was retrieved then this is the apriori.
        Shape (numret, 100)
    ('retrieved', 'aero') : 2d numpy array
        Retrieved aerosol number density.  If no aerosol was retrieved then this is the apriori
        shape (numret, 100)
    ('retrieved', 'aerosol_extincton') : 2d numary array
        Same as retrieved aerosol but in units of extinction rather than number density. Shape (numret, 100)
    ('retrieved', 'albedo') : 2d numpy array
        Retrieved albedo.  If no albedo was retrieved then this is the apriori
    ('ret_opt') : dict
        A copy of all of the retrieval options
    ('simulated', 'radiance') : 2d numpy array
        (los, wavel) the full spectrum of radiances calculated at the end of the retrieval.  Only stored
        if ret_opt['model_spectrum'] is True.  Same wavelengths as ('omps_measurement', 'wavel_nm')

    """
    def __init__(self, orbitnumber):
        self.orbitnumber = orbitnumber
        self.ret_opt = self.retrieval_options_default()

        # Maps the user retrieval order strings to internal functions
        self.ret_func_dict = {'ozone': lambda a, b: self.retrieve_ozone(a, b),
                              'aerosol': lambda a, b: self.retrieve_aerosol(a, b),
                              'albedo': lambda a, b: self.retrieve_albedo(a, b),
                              }

        self.output = dict()

    @staticmethod
    def cloud_height_from_omps_model(omps_model: omps.OMPS) -> float:
        radiance = xr.DataArray(omps_model.rad,
                                dims=('tangent_height', 'wavelength'),
                                coords={'wavelength': omps_model.wavel_nm,
                                        'tangent_height': omps_model.tanalts / 1000})

        return cd.cloud_alt_km(radiance)

    def make_meas_inst(self):
        self.filename = self.l1_filename_for_orbit(self.orbitnumber)
        alllos = []
        allobs = []
        allmjd = []
        allrad = []
        scanslices = []
        alllat = []
        alllon = []
        allsza = []

        curidx = 0
        scanidx = []

        # Start by picking a scan in the file and pulling out the wavelengths
        # TODO: Pick this scan better
        inst_sim = omps.OMPS(self.filename, 80, 1)
        wavel_nm = inst_sim.wavel_nm[10:240]
        psf = inst_sim.bandpass[10:240]

        self.cloud_height_km = []

        for idx, image in enumerate(self.ret_opt['images']):
            # Put a try statement here incase the OMPS L1 file doesnt have
            # data for a specified index
            try:
                inst_sim = omps.OMPS(self.filename, image, 1)
                los, obs, mjd = inst_sim.radtran_geometry()

                if inst_sim.sza > self.ret_opt['max_sza']:
                    continue
                self.cloud_height_km.append(self.cloud_height_from_omps_model(inst_sim))

                rad = sp.interpolate.interp1d(inst_sim.wavel_nm, inst_sim.rad, axis=1)(wavel_nm)

                alllos.append(los)
                allobs.append(obs)
                allmjd.append(mjd)
                allsza.append(inst_sim.sza)
                allrad.append(rad)
                scanslices.append(np.arange(curidx, curidx+len(los)))
                curidx += len(los)
                alllat.append(inst_sim.lat)
                alllon.append(inst_sim.lon)
                scanidx.append(image)
            except Exception as e:
                logging.exception(e)
        self.cloud_height_km = np.array(self.cloud_height_km)
        alllos = np.array(alllos).reshape((-1, 3))
        allobs = np.array(allobs).reshape((-1, 3))
        allmjd = np.array(allmjd).reshape((-1, 1))[:]
        allrad = np.array(allrad).reshape((-1, len(wavel_nm)))

        # Put the data in an ideal spectrograph container
        meas_inst = idealspectrograph.IdealSpectrograph(wavel_nm, alllos, allobs, allmjd)
        meas_inst.psf = psf
        meas_inst.rad = allrad
        # Assume 1% error in the radiance measurements
        meas_inst.raderror = allrad * 0.01

        self.output['omps_geometry', 'image_index'] = scanidx
        self.output['omps_geometry', 'sza'] = allsza
        self.output['omps_geometry', 'mjd'] = allmjd
        self.output['omps_geometry', 'latitude'] = alllat
        self.output['omps_geometry', 'longitude'] = alllon

        self.output['omps_measurement', 'radiance'] = allrad
        self.output['omps_measurement', 'wavel_nm'] = wavel_nm
        self.output['omps_measurement', 'psf'] = psf

        try:
            self.trop_alt = tropopause_alt(self.orbitnumber, self.ret_opt['omps_l1_anc_folder'])[scanidx]
        except:
            self.trop_alt = None
        
        self.output['omps_geometry', 'trop_alt'] = self.trop_alt
        self.output['omps_geometry', 'cloud_alt'] = self.cloud_height_km * 1000

        return meas_inst, scanslices

    def make_meas_inst_noaa(self):
        alllos = []
        allobs = []
        allmjd = []
        allrad = []
        scanslices = []
        alllat = []
        alllon = []
        allsza = []
        alltrop = []

        curidx = 0
        scanidx = []

        self.cloud_height_km = []

        for l1gfile, ancfile in zip(self.ret_opt['l1g_input_files'], self.ret_opt['l1_anc_input_files']):
            h5_file = h5py.File(l1gfile)
            anc_file = h5py.File(ancfile)

            num_images = h5_file['GEOLOCATION_DATA/Latitude_25km'].shape[0]


            for image in range(num_images):
                inst_sim = omps.OMPS(l1gfile, image, 1)
                wavel_nm = inst_sim.wavel_nm[10:240]
                psf = inst_sim.bandpass[10:240]
                los, obs, mjd = inst_sim.radtran_geometry()

                if inst_sim.sza > self.ret_opt['max_sza']:
                    continue
                self.cloud_height_km.append(self.cloud_height_from_omps_model(inst_sim))

                rad = sp.interpolate.interp1d(inst_sim.wavel_nm, inst_sim.rad, axis=1)(wavel_nm)

                alltrop.append(anc_file['GRIDDED_DATA/TropopauseAltitude'][image, 1]*1000)

                alllos.append(los)
                allobs.append(obs)
                allmjd.append(mjd)
                allsza.append(inst_sim.sza)
                allrad.append(rad)
                scanslices.append(np.arange(curidx, curidx + len(los)))
                curidx += len(los)
                alllat.append(inst_sim.lat)
                alllon.append(inst_sim.lon)
                scanidx.append(image)

        self.cloud_height_km = np.array(self.cloud_height_km)
        alllos = np.array(alllos).reshape((-1, 3))
        allobs = np.array(allobs).reshape((-1, 3))
        allmjd = np.array(allmjd).reshape((-1, 1))[:]
        allrad = np.array(allrad).reshape((-1, len(wavel_nm)))

        # Put the data in an ideal spectrograph container
        meas_inst = idealspectrograph.IdealSpectrograph(wavel_nm, alllos, allobs, allmjd)
        meas_inst.psf = psf
        meas_inst.rad = allrad
        # Assume 1% error in the radiance measurements
        meas_inst.raderror = allrad * 0.01

        self.output['omps_geometry', 'image_index'] = scanidx
        self.output['omps_geometry', 'sza'] = allsza
        self.output['omps_geometry', 'mjd'] = allmjd
        self.output['omps_geometry', 'latitude'] = alllat
        self.output['omps_geometry', 'longitude'] = alllon

        self.output['omps_measurement', 'radiance'] = allrad
        self.output['omps_measurement', 'wavel_nm'] = wavel_nm
        self.output['omps_measurement', 'psf'] = psf

        self.trop_alt = np.array(alltrop)

        self.output['omps_geometry', 'trop_alt'] = self.trop_alt
        self.output['omps_geometry', 'cloud_alt'] = self.cloud_height_km * 1000

        return meas_inst, scanslices

    def HR_options(self):
        options = opt.HROptions()
        options.add_option('numordersofscatter', self.ret_opt['scatterorder'])
        options.add_option('raytracingshells', 1000)
        options.add_option('numdiffuseprofilesinplane', self.ret_opt['num_diffuse_profiles'])
        options.add_option('solarraytracingshells', 1000)
        options.add_option('integrationtechnique', 1)
        options.add_option('maxopticaldepthofcell', 0.1)
        options.add_option('minextinctionratioofcell', 0.9)
        options.add_option('wfprecision', 1)
        options.add_option('numthreads', 0)
        options.add_option('solartablenumcossza', 361)
        options.add_option('wfwidths', np.ones(100)*2000)
        diff_heights = np.concatenate(([10**-6], np.linspace(500, 24500, 13), np.linspace(29500, 99500, 15)))
        options.add_option('manualdiffuseheights', diff_heights)
        return options

    def l1_filename_for_orbit(self, orbitnumber):

        files = listdir(self.ret_opt['omps_l1_folder'])

        for f in files:
            m = re.search('o([0-9]*)', f)
            if m is not None:
                if len(m.group(1)) > 0:
                    orbit = int(m.group(1))
                    if orbit == orbitnumber:
                        return self.ret_opt['omps_l1_folder'] + f
        raise Exception('Orbit does not exist in folder')

    @staticmethod
    def retrieval_options_default():
        ret_opt = dict()
        ret_opt['max_sza'] = 88
        ret_opt['scatterorder'] = 30
        ret_opt['retrieval_order'] = ['albedo', 'aerosol', 'ozone', 'albedo', 'ozone']
        ret_opt['max_iter'] = [5, 5, 5, 2, 2]
        ret_opt['images'] = range(0, 180)
        ret_opt['verbose_output'] = False
        ret_opt['maxdifference_seconds'] = 100
        ret_opt['omps_l1_folder'] = r"\\utls\\users\\data\\OMPS\\OMPS_L1\\"
        ret_opt['omps_l1_anc_folder'] = r"\\utls\\users\\data\\OMPS\\OMPS_ANC\\"
        ret_opt['omps_l2_folder'] = r"\\utls\\users\\data\\OMPS\\OMPS_L2\\"
        ret_opt['mart_folder'] = r"\\utls\\users\\scratch\\daniel\\"
        ret_opt['air_climatology'] = 'omps_anc'
        ret_opt['ozone_apriori_type'] = 'labow'
        ret_opt['aerosol_apriori_type'] = 'default'
        ret_opt['grid_type'] = 'uniform'
        ret_opt['grid_down_factor'] = 1.0
        ret_opt['tropopause_lowerbound'] = True
        ret_opt['cloud_lowerbound'] = True
        ret_opt['ozone_apriori_scaling'] = True
        ret_opt['aerosol_smooth_factor'] = 0.0
        ret_opt['num_diffuse_profiles'] = 5
        ret_opt['albedo_smooth_factor'] = 10
        ret_opt['model_spectrum'] = False
        ret_opt['o3_horiz_tikh_factor'] = 2.0
        ret_opt['aero_horiz_tikh_factor'] = 4.0
        ret_opt['o3_vertical_tikh_factor'] = 2.0
        ret_opt['aero_vertical_tikh_factor'] = 4.0
        ret_opt['retrieval_method'] = 'lm'
        ret_opt['ozone_lm_damping'] = 0.1
        ret_opt['aerosol_lm_damping'] = 0.01
        ret_opt['save_averaging_kernel'] = False
        ret_opt['ozone_config_file'] = os.path.join(os.path.dirname(os.path.realpath(tomospecies.__file__)), 'ozone_default.yaml')
        ret_opt['l1g_input_files'] = None
        ret_opt['l1_anc_input_files'] = None
        return ret_opt

    def retrieve_aerosol(self, maxiter, do_error_analysis=True):
        self.method.maxiter = maxiter
        self.method.do_error_analysis = do_error_analysis
        self.atmo_ret.set_wf_species(self.optprop_aer)
        self.method.damping = self.ret_opt['aerosol_lm_damping']
        if type(self.method) is rodgers.Rodgers:
            self.method.set_apriori(self.aero_apriori_x)
            self.method.calc_averaging_kernel = self.ret_opt['save_averaging_kernel']
        self.ret.retrieve_species(self.species_aero, self.method, do_error=True)

        ymeas, e = self.species_aero.measurement_vector(self.ret.meas_inst)
        yret, e = self.species_aero.measurement_vector(self.ret.ret_inst)

        rad_ret = self.ret.ret_inst.rad
        wavel_ret = self.ret.ret_inst.wavel_nm

        debug = dict()
        debug['ymeas'] = ymeas
        debug['yret'] = yret
        debug['rad_ret'] = rad_ret
        debug['wavel_ret'] = wavel_ret

        if type(self.method) is rodgers.Rodgers and do_error_analysis:
            self.output['retrieved', 'aerosol_numden_error'] = self.species_aero.propagate_covariance(self.method.output['measurement_error_covariance'])
            debug['averaging_kernel'] = self.method.A
            debug['chi_squared'] = self.method.chi_squared
        else:
            self.output['retrieved', 'aerosol_numden_error'] = np.ones(np.shape(self.species_aero.profile)) * np.nan

        return self.species_aero.output_data()[0], debug

    def retrieve_ozone(self, maxiter, do_error_analysis=True):
        self.method.maxiter = maxiter
        self.method.do_error_analysis = do_error_analysis
        self.atmo_ret.set_wf_species(self.optprop_o3)
        self.method.damping = self.ret_opt['ozone_lm_damping']

        if type(self.method) is rodgers.Rodgers:
            self.method.set_apriori(self.o3_apriori_x)
            self.method.calc_averaging_kernel = self.ret_opt['save_averaging_kernel']

        self.ret.retrieve_species(self.species_o3, self.method, do_error=True)

        ymeas, e = self.species_o3.measurement_vector(self.ret.meas_inst)
        yret, e = self.species_o3.measurement_vector(self.ret.ret_inst)

        rad_ret = self.ret.ret_inst.rad
        wavel_ret = self.ret.ret_inst.wavel_nm

        debug = dict()
        debug['ymeas'] = ymeas
        debug['yret']  = yret
        debug['rad_ret'] = rad_ret
        debug['wavel_ret'] = wavel_ret

        if type(self.method) is rodgers.Rodgers and do_error_analysis:
            self.output['retrieved', 'ozone_numden_error'] = self.species_o3.propagate_covariance(self.method.output['measurement_error_covariance'])
            debug['averaging_kernel'] = self.method.output['averaging_kernel']
            debug['chi_squared'] = self.method.output['residual'][-1]
        else:
            self.output['retrieved', 'ozone_numden_error'] = np.ones(np.shape(self.species_o3.profile)) * np.nan

        return self.species_o3.output_data()[0], debug

    def retrieve_albedo(self, maxiter, do_error_analysis=True):
        self.atmo_ret.set_wf_species(self.optprop_o3)
        self.albedo_method.maxiter = maxiter
        self.ret.retrieve_species(self.species_albedo, self.albedo_method)

        ymeas, e = self.species_albedo.measurement_vector(self.ret.meas_inst)
        yret, e = self.species_albedo.measurement_vector(self.ret.ret_inst)

        rad_ret = self.ret.ret_inst.rad
        wavel_ret = self.ret.ret_inst.wavel_nm

        debug = dict()
        debug['ymeas'] = ymeas
        debug['yret'] = yret
        debug['rad_ret'] = rad_ret
        debug['wavel_ret'] = wavel_ret
        return self.species_albedo.output_data(), debug

    def apriori_ozone(self, meas_inst):
        optprop_o3 = ozone.convolved_dbm_optical_property(meas_inst.wavel_nm, meas_inst.psf)
        guid_o3 = skif.SKCLIMATOLOGY_O3_CM3
        data_dir = uutil.data_folder()
        if self.ret_opt['ozone_apriori_type'] == 'omps_l2':
            try:
                data = omps_l2.load_ozone_orbit(self.orbitnumber, 2)
            except:
                logging.warning('OMPS_L2 ozone file does not exist for this orbit')
            o3 = data['o3']
            heights = data['height'] * 1000  # in m

            # TODO: This needs to be fixed for when the grids are not the same size
            prof_o3 = np.zeros((meas_inst.radtran.retrieval_grid.numretprof, len(meas_inst.radtran.retrieval_grid.altitudes)))
            # for each profile, interpolate onto log space
            for idx, image_idx in enumerate(self.output['omps_geometry', 'image_index']):
                good = ~np.isnan(o3[image_idx, :]) & (o3[image_idx, :] > 0)
                if np.sum(good) == 0:
                    prof_o3[idx, :] = np.loadtxt(data_dir + r'labow.txt').transpose()
                else:
                    f_i = InterpolatedUnivariateSpline(heights[good], np.log(o3[image_idx,good]), k=1)
                    prof_o3[idx, :] = np.exp(f_i(meas_inst.radtran.retrieval_grid.altitudes))
                    temp = np.loadtxt(data_dir + r'labow.txt').transpose()
                    prof_o3[idx, ~good] = temp[~good]

        elif self.ret_opt['ozone_apriori_type'] == 'constant':

            prof_o3 = np.loadtxt(data_dir + r'labow.txt').transpose()
            prof_o3 = np.tile(prof_o3, (meas_inst.radtran.retrieval_grid.numretprof, 1))

        elif self.ret_opt['ozone_apriori_type'] == 'mart':
            o3 = np.load(self.ret_opt['mart_folder'] + 'ozone_12112.npy')
            image_idx = self.output['omps_geometry', 'image_index']
            prof_o3 = o3[:, image_idx].transpose()

            proflat = self.output['omps_geometry', 'latitude']
            proflon = self.output['omps_geometry', 'longitude']
            profheight = np.linspace(500, 99500, 100)

            prof_o3 = meas_inst.radtran.retrieval_grid.interpolate_onto_grid(proflon, proflat, profheight, prof_o3)

        elif self.ret_opt['ozone_apriori_type'] == 'labow':
            lat, lon, alt = geometry.lat_lon_alts(meas_inst.radtran.retrieval_grid.retprofileloc)
            mjd = meas_inst.radtran.retrieval_grid.retprofilemjd

            prof_o3 = np.zeros((meas_inst.radtran.retrieval_grid.numretprof, len(meas_inst.radtran.retrieval_grid.altitudes)))

            labow = skif.ISKClimatology('O3LABOW')
            for idx in range(meas_inst.radtran.retrieval_grid.numretprof):
                prof_o3[idx,:] = utility.height_profile_from_clim(labow, guid_o3, meas_inst.radtran.retrieval_grid.altitudes, lat[idx], lon[idx], mjd[idx])

        self.output['apriori', 'o3'] = copy.copy(prof_o3)

        return prof_o3, optprop_o3, guid_o3

    def omps_anc_background(self, grid, atmo):
        num_image = len(self.output['omps_geometry', 'image_index'])
        t_prof = np.zeros((num_image, 102))
        p_prof = np.zeros((num_image, 102))
        n_prof = np.zeros((num_image, 102))
        for idx, image_idx in enumerate(self.output['omps_geometry', 'image_index']):
            alt, t_prof[idx, :], p_prof[idx, :], n_prof[idx, :] = background.omps_anc_profile(self.orbitnumber, image_idx, 1, folder=self.ret_opt['omps_l1_anc_folder'])
        proflat = self.output['omps_geometry', 'latitude']
        proflon = self.output['omps_geometry', 'longitude']

        t_prof = grid.interpolate_onto_grid(proflon, proflat, alt, t_prof, False)
        p_prof = grid.interpolate_onto_grid(proflon, proflat, alt, p_prof, False)
        n_prof = grid.interpolate_onto_grid(proflon, proflat, alt, n_prof, False)

        self.output['retrieval_geometry', 'airtemperature'] = t_prof
        self.output['retrieval_geometry', 'airpressure'] = p_prof
        self.output['retrieval_geometry', 'airnumberdensity'] = n_prof
        self.output['retrieval_geometry', 'airalt'] = alt

        atmo.set_user_defined_background(grid, skif.ISKOpticalProperty('RAYLEIGH'), alt, t_prof, p_prof, n_prof)
    
    def omps_anc_background_noaa(self, grid, atmo):
        num_image = len(self.output['omps_geometry', 'image_index'])
        t_prof = np.zeros((num_image, 102))
        p_prof = np.zeros((num_image, 102))
        n_prof = np.zeros((num_image, 102))

        c = 0
        for anc_file in self.ret_opt['l1_anc_input_files']:
            anc_h5 = h5py.File(anc_file)

            num_image = anc_h5['/GEOLOCATION_DATA/Latitude'].shape[0]

            for idx in range(num_image):
                alt, t_prof[c, :], p_prof[c, :], n_prof[c, :] = background.omps_anc_profile(0, idx, 1, filename=anc_file)
                c += 1

        proflat = self.output['omps_geometry', 'latitude']
        proflon = self.output['omps_geometry', 'longitude']

        t_prof = grid.interpolate_onto_grid(proflon, proflat, alt, t_prof, False)
        p_prof = grid.interpolate_onto_grid(proflon, proflat, alt, p_prof, False)
        n_prof = grid.interpolate_onto_grid(proflon, proflat, alt, n_prof, False)

        self.output['retrieval_geometry', 'airtemperature'] = t_prof
        self.output['retrieval_geometry', 'airpressure'] = p_prof
        self.output['retrieval_geometry', 'airnumberdensity'] = n_prof
        self.output['retrieval_geometry', 'airalt'] = alt

        atmo.set_user_defined_background(grid, skif.ISKOpticalProperty('RAYLEIGH'), alt, t_prof, p_prof, n_prof)

    def apriori_aerosol(self, meas_inst):
        data_dir = uutil.data_folder()
        clim_aer, optprop_aer, guid_aer = aerosol.from_text_file(data_dir + r'aerosol_apriori.txt', mode_radius=[0.08, 0.08])
        if self.ret_opt['aerosol_apriori_type'] == 'default':
            prof = np.loadtxt(data_dir + r'aerosol_apriori.txt')
            prof_aero = np.interp(meas_inst.radtran.retrieval_grid.altitudes, prof[:, 0], prof[:, 1]).transpose()
            prof_aero = np.tile(prof_aero, (meas_inst.radtran.retrieval_grid.numretprof, 1))
        elif self.ret_opt['aerosol_apriori_type'] == 'mart':
            aero = np.load(self.ret_opt['mart_folder'] + 'aer_12112.npy')
            image_idx = self.output['omps_geometry', 'image_index']
            prof_aero = aero[:, image_idx].transpose()

            proflat = self.output['omps_geometry', 'latitude']
            proflon = self.output['omps_geometry', 'longitude']
            profheight = np.linspace(500, 99500, 100)

            prof_aero = meas_inst.radtran.retrieval_grid.interpolate_onto_grid(proflon, proflat, profheight, prof_aero)

        prof_aero[prof_aero <= 0] = 1e-5
        self.output['apriori', 'aero'] = copy.copy(prof_aero)
        return prof_aero, optprop_aer, guid_aer

    def retrieve_orbit(self):
        if self.ret_opt['retrieval_method'].lower() == 'rodgers':
            self.method = rodgers.Rodgers(do_sparse_linalg=True, damping=0.01)
        elif self.ret_opt['retrieval_method'].lower() == 'lm':
            self.method = lsq.LevenbergMarquardt(damping=0.01, maxiter=7)
        else:
            logging.error('Invalid Retrieval Method')
            raise ValueError('Invalid Retrieval Method')

        if self.ret_opt['l1g_input_files'] is None:
            meas_inst, scanslices = self.make_meas_inst()
        else:
            meas_inst, scanslices = self.make_meas_inst_noaa()

        atmo_ret = tomography.atmosphere.atmo_2d.AtmosphereTwoDim()
        model_ret = tomography.sasktran_2d.SasktranHRTwoDim(atmo_ret, self.HR_options(), maxdifference_seconds=self.ret_opt['maxdifference_seconds'])
        ret_inst  = idealspectrograph.IdealSpectrograph([], meas_inst.los, meas_inst.observer, meas_inst.mjd)
        ret_inst.setup_radtran_model(model_ret)
        meas_inst.radtran = ret_inst.radtran
        model_ret.setup_geometry(scanslices, placementtype=self.ret_opt['grid_type'], numinbetween=self.ret_opt['grid_down_factor'])

        # Setup the apriori atmosphere
        if self.ret_opt['air_climatology'] == 'ecmwf':
            ecmwf = skif.ISKClimatology('ECMWF')
            atmo_ret.add_species(ecmwf,
                                 skif.ISKOpticalProperty('Rayleigh'),
                                 skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3 )
            atmo_ret.backgroundclim = ecmwf
        if self.ret_opt['air_climatology'] == 'omps_anc':
            if self.ret_opt['l1_anc_input_files'] is None:
                self.omps_anc_background(model_ret.retrieval_grid, atmo_ret)
            else:
                self.omps_anc_background_noaa(model_ret.retrieval_grid, atmo_ret)

        prof_aero, optprop_aer, guid_aer = self.apriori_aerosol(meas_inst)
        prof_o3, optprop_o3, guid_o3 = self.apriori_ozone(meas_inst)

        atmo_ret.add_user_defined_twodim(model_ret.retrieval_grid, guid_o3 , optprop_o3, prof_o3, 'tomographic_ozone')
        atmo_ret.add_user_defined_twodim(model_ret.retrieval_grid, guid_aer , optprop_aer, prof_aero, 'tomographic_aerosol')
        twodim_albedo = np.ones(model_ret.retrieval_grid.numretprof) * 0.3
        atmo_ret.set_twodim_albedo(twodim_albedo)

        clim_no2, optprop_no2, guid_no2 = no2.pratmo()
        optprop_no2 = no2.convolved_vandaele_optical_property(meas_inst.wavel_nm, meas_inst.psf)
        atmo_ret.add_species(clim_no2, optprop_no2, guid_no2)

        ret_lat, ret_lon, ret_alts = geometry.lat_lon_alts(model_ret.retrieval_grid.retprofileloc)

        self.output['retrieval_geometry', 'latitude'] = ret_lat
        self.output['retrieval_geometry', 'longitude'] = ret_lon
        self.output['retrieval_geometry', 'altitudes'] = model_ret.retrieval_grid.altitudes
        self.output['retrieval_geometry', 'location_xyz'] = model_ret.retrieval_grid.retprofileloc
        self.output['retrieval_geometry', 'mjd'] = model_ret.retrieval_grid.retprofilemjd
        self.output['retrieval_geometry', 'scanslices'] = model_ret.scanslices


        # Setup our retrieval classes
        self.species_aero = tomospecies.TomographicAerosol(atmo_ret, True, horiz_tikh_factor=self.ret_opt['aero_horiz_tikh_factor'],
                                                           vertical_tikh_factor=self.ret_opt['aero_vertical_tikh_factor'])
        self.species_o3 = tomospecies.TomographicOzone(atmo_ret, self.ret_opt['ozone_apriori_scaling'],
                                                       horiz_tikh_factor=self.ret_opt['o3_horiz_tikh_factor'],
                                                       vertical_tikh_factor=self.ret_opt['o3_vertical_tikh_factor'],
                                                       config_file=self.ret_opt['ozone_config_file'])

        try:
            self.o3_apriori_x = copy.copy(self.species_o3.get_x())
        except AttributeError:
            # Not doing ozone retrieval, do nothing
            pass
        try:
            self.aero_apriori_x = copy.copy(self.species_aero.get_x())
        except AttributeError:
            # Not doing aerosol, dont need this
            pass

        self.species_albedo = tomoalbedo.TomographicAlbedo(atmo_ret, smooth_factor = self.ret_opt['albedo_smooth_factor'])
        self.ret = core.RetrievalLimb(meas_inst, ret_inst)

        if self.ret_opt['tropopause_lowerbound'] and self.trop_alt is not None:
            trop_alt = self.trop_alt

            self.species_o3.lowalt = trop_alt-2000
            self.species_aero.lowalt = (trop_alt-6000)  # Convert to _m

        if self.ret_opt['cloud_lowerbound']:
            self.species_o3.lowalt = np.maximum(self.species_o3.lowalt, self.cloud_height_km*1000)
            self.species_aero.lowalt = np.maximum(self.species_aero.lowalt, self.cloud_height_km*1000)

        self.albedo_method = lsq.MART()

        # Variables that the retrieval functions need
        self.atmo_ret = atmo_ret
        self.optprop_o3 = optprop_o3
        self.optprop_aer = optprop_aer
        for idx, (spec, o) in enumerate(zip(self.ret_opt['retrieval_order'], self.ret_opt['max_iter'])):
            # True if its the last pass for a given species
            do_error = not any((np.array(self.ret_opt['retrieval_order']) == spec)[(idx+1):])
            data = self.ret_func_dict[spec](o, do_error)
            self.output['debug_' + spec] = data

        numret = len(ret_lat)

        try:
            self.output['retrieved', 'o3'] = self.species_o3.output_data()[0]
            if not self.output['retrieved', 'o3'].any():
                self.output['retrieved', 'o3'] = np.zeros((numret, 100))

            self.output['retrieved', 'o3_RTM'] = copy.copy(self.output['retrieved', 'o3'])
            high_idx =  np.nonzero(self.species_o3.highalt < self.species_o3.tanalts)[0][0]
            self.output['retrieved', 'o3'][:, high_idx:] = np.nan

        except:
            self.output['retrieved', 'o3'] = np.zeros((numret, 100))
            self.output['retrieved', 'o3_RTM'] = np.zeros((numret, 100))
        try:
            self.output['retrieved', 'aero'] = self.species_aero.output_data()[0]

            if not self.output['retrieved', 'aero'].any():
                self.output['retrieved', 'aero'] = np.zeros((numret, 100))

            # NaN the aerosol profile above the highest alt
            self.output['retrieved', 'aero_RTM'] = copy.copy(self.output['retrieved', 'aero'])
            low_norm_idx = np.array([s[0] for s in self.species_aero.scan_alt_norm_idx])
            low_norm_idx_on_grid = model_ret.retrieval_grid.interpolate_onto_grid(self.output['omps_geometry', 'longitude'],
                                                                                  self.output['omps_geometry', 'latitude'],
                                                                                  [],
                                                                                  low_norm_idx, do_height_interp=False)

            for idx in range(len(low_norm_idx_on_grid)):
                self.output['retrieved', 'aero'][idx, int(np.ceil(low_norm_idx_on_grid)[idx]):] = np.nan


            self.output['retrieval_geometry', 'aero_norm'] = self.species_aero.scan_alt_norm_idx
        except:
            self.output['retrieved', 'aero'] = np.zeros((numret, 100))
            self.output['retrieved', 'aero_RTM'] = np.zeros((numret, 100))
        try:
            self.output['retrieved', 'albedo'] = self.species_albedo.output_data()
            if not self.output['retrieved', 'albedo'].any():
                self.output['retrieved', 'albedo'] = np.zeros(numret)
        except:
            self.output['retrieved', 'albedo'] = np.zeros(numret)

        self.output['ret_opt'] = self.ret_opt

        if self.ret_opt['model_spectrum']:
            self.output['simulated', 'radiance'] = np.zeros(np.shape(self.output['omps_measurement', 'radiance']))
            for idx, w in enumerate(self.output['omps_measurement', 'wavel_nm']):
                ret_inst.set_wavelengths([w])
                ret_inst.calculate_radiance()
                self.output['simulated', 'radiance'][:, idx] = copy.copy(np.squeeze(ret_inst.rad))

        # Record the tropopause altitude and cloud altitudes on the retrieval grid
        self.output['retrieved', 'trop_alt'] = meas_inst.radtran.retrieval_grid.interpolate_onto_grid(
            self.output['omps_geometry', 'longitude'], self.output['omps_geometry', 'latitude'], [], self.trop_alt
        )

        self.output['retrieved', 'cloud_alt'] = meas_inst.radtran.retrieval_grid.interpolate_onto_grid(
            self.output['omps_geometry', 'longitude'], self.output['omps_geometry', 'latitude'], [], self.cloud_height_km * 1000
        )

        # Calculate the aerosol extinction as well
        aer_ext = np.zeros(np.shape(self.output['retrieved', 'aero']))

        if self.species_aero.wavel_nm is not None:
            for idx in range(len(self.output['retrieval_geometry', 'latitude'])):
                aer_ext[idx, :] = utility.number_density_to_extinction(self.output['retrieval_geometry', 'altitudes'],
                                                                       self.output['retrieved', 'aero'][idx, :],
                                                                       self.species_aero.wavel_nm[0],
                                                                       self.optprop_aer,
                                                                       )
            try:
                aer_ext_error = np.zeros(np.shape(self.output['retrieved', 'aerosol_numden_error']))
                for idx in range(len(self.output['retrieval_geometry', 'latitude'])):
                    aer_ext_error[idx, :] = utility.number_density_to_extinction(
                        self.output['retrieval_geometry', 'altitudes'],
                        self.output['retrieved', 'aerosol_numden_error'][idx, :],
                        self.species_aero.wavel_nm[0],
                        self.optprop_aer,
                        )
                self.output['retrieved', 'aerosol_extinction_error'] = aer_ext_error
            except:
                pass

            self.output['retrieved', 'aerosol_wavelength'] = self.species_aero.wavel_nm[0]
        else:
            self.output['retrieved', 'aerosol_wavelength'] = np.nan

        self.output['retrieved', 'aerosol_extinction'] = aer_ext

        self.output['processing', 'date'] = time.strftime("%c")
        self.output['processing', 'version'] = '1.0.2'

        if not ('retrieved', 'ozone_numden_error') in self.output:
            self.output['retrieved', 'ozone_numden_error'] = np.ones((numret, 100))*np.nan

        if not ('retrieved', 'aerosol_numden_error') in self.output:
            self.output['retrieved', 'aerosol_numden_error'] = np.ones((numret, 100))*np.nan
            self.output['retrieved', 'aerosol_extinction_error'] = np.ones((numret, 100))*np.nan


        return self.output
