import numpy as np
from scipy import interpolate
import radtran.geometry as geometry
import logging
import scipy as sp


def correct_for_earths_rotation(tp: np.ndarray, mjd: np.ndarray, refmjd: float) -> np.ndarray:
    """
    Takes a set of tangent points, corresponding mjds and rotates them by
    the Earth's rotation.    If you take a set of satellite tangent points
    and mjd's and apply this function, the end result should be a plane
    
    Parameters
    ----------
    tp: np.ndarray
        (n, 3) of tangent points
    mjd: np.ndarray
        (n,) modified julian dates
    refmjd : float
        The reference mjd where no rotation is applied
    
    Returns
    -------
    2d numpy array
        Shape (n,3) of rotated tangent points
    """
    omega = 2 * np.pi

    mjddiff = mjd - refmjd
    rottp = np.zeros(np.shape(tp))

    for idx in range(np.size(mjddiff)):
        theta = omega * mjddiff[idx]
        rot_mat = np.array([[np.cos(theta), -1 * np.sin(theta), 0], [np.sin(theta), np.cos(theta), 0], [0, 0, 1]])

        rottp[idx, :] = rot_mat.dot(tp[idx, :])

    return rottp


def project_onto_plane(tp: np.ndarray, normal: np.ndarray) -> np.ndarray:
    """
    Takes a list of tangent point vectors and returns their value projected onto a plane specified by normal.
    Note that the projected value returned has the same magnitude as the original input tangent point.  The only
    thing different is the direction
    Parameters
    ----------
    tp: np.ndarray
        (:, 3) array of tangent points
    normal: np.ndarray
        (3,) normal unit vector for the plane

    Returns
    -------
    projected: np.ndarray
        The tangent points projected into the plane, same size as tp
    """
    projected = np.zeros(np.shape(tp))
    for idx, t in enumerate(tp):
        n = np.linalg.norm(t)
        projected[idx, :] = t - np.dot(t, normal) * normal
        projected[idx, :] *= n / np.linalg.norm(projected[idx, :])

    return projected


def rotation_matrix(axis: np.ndarray, theta: float) -> np.ndarray:
    """
    Calculates the rotation matrix for an arbitrary rotation axis and angle
    
    Parameters
    ----------
    axis : np.ndarray
        Shape (3,), unit vector of rotation axis
    theta : scalar
        Angle in radians to rotate by
    
    Returns
    -------
    2d numpy array
        Rotation matrix with shape (3,3)
    """
    axis = np.asarray(axis)
    axis = axis / np.sqrt(np.dot(axis, axis))
    a = np.cos(theta / 2)
    b, c, d = -axis * np.sin(theta / 2)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])


class TwoDimGrid(object):
    """
    Class which constructs a two dimensional grid out of a set of tangent
    points and look directions.  The core idea is that when the earths rotation
    is corrected for, the tangent points should form a plane.  A uniform in angle
    grid is constructed in this plane.  The approximation made is that over the
    course of one radiative transfer calculation, the actual grid is almost the
    same as this rotation corrected grid.
    
    The inputs to this function should be the low resolution values
    
    Parameters
    ----------
    tp : list or 2d numpy array
        Tangent points, should be convertible to a numpy array of shape(n,3)
    look : list or 2d numpy array
        Look directions, should be convertible to a numpy array of shape(n,3)
    mjd : List or numpy array
        Modified julian dates
    scanslices : list
        Each list element should be a range indicating which line of sights
        belong to which scan. e.g. scanslices[0] could be range(0,60) indicating
        that lines of sight 0-60 are in the first scan
    numinbetween : scalar, optional
        The ratio of scans / retrieval profiles if placementtype is set to uniform
    cutoff : scalar, optional
        Angle in degrees to cutoff of each side of the grid if placementype is
        set to uniform
    placementtype : string, optional
        If set to 'scan', retrieval profiles are placed at the average tp of
        every scan.  If set to uniform, a uniform grid is constructed between
        the first and last scan
        
    """

    def __init__(self, tp, look, mjd, scanslices, numinbetween=1.0, cutoff=0, placementtype='scan'):
        self.tp = np.array(tp)  # Tangent points as a list of numpy arrays
        self.look = np.array(look)  # Unit look vectors as a list of numpy arrays
        self.mjd = np.array(mjd)  # Array of MJD's
        self.scanslices = np.array(scanslices)  # Array of indicies indicating where each scan happens
        self.numinbetween = numinbetween  # The number of scans / ret profiles
        self.cutoff = np.deg2rad(cutoff)

        # Find the average tangent point of each scan, this could be changed to
        # Find the 30km tangent point or something similar
        self.averagescantp = []
        self.averagescanlook = []
        self.averagescanmjd = []
        self.altitudes = np.linspace(500, 99500, 100)

        for s in scanslices:
            self.averagescantp.append(np.mean(self.tp[s], axis=0))
            self.averagescanlook.append(np.mean(self.look[s], axis=0))
            self.averagescanmjd.append(np.mean(self.mjd[s]))

        self.averagescantp = np.asarray(self.averagescantp)
        self.averagescanlook = np.asarray(self.averagescanlook)
        self.averagescanmjd = np.asarray(self.averagescanmjd)

        self.averagescanangles = None
        self.xaxis = None
        self.yaxis = None
        self.numretprof = None
        self.angles = None
        self.retprofilemjd = None
        self.retprofileavglook = None
        self.retprofileloc = None

        self.placementtype = placementtype

        self.make_retrieval_profile_locations()

    def make_retrieval_profile_locations(self):
        """
        Takes the measurement grid and constructs the retrieval grid.
        """
        rottp = correct_for_earths_rotation(self.averagescantp, self.averagescanmjd, self.averagescanmjd[0])
        localangles = np.zeros(np.size(self.averagescanmjd))
        xaxis = rottp[0, :] / np.linalg.norm(rottp[0, :])
        locallook = (rottp[np.min([10, np.size(self.averagescanmjd)])] - rottp[0, :])
        locallook /= np.linalg.norm(locallook)
        normal = np.cross(locallook, xaxis)
        normal /= np.linalg.norm(normal)
        yaxis = np.cross(xaxis, normal)
        yaxis /= np.linalg.norm(yaxis)
        twopiaddition = 0
        for idx in range(1, np.size(localangles)):
            localangles[idx] = np.arctan2(yaxis.dot(rottp[idx, :]), xaxis.dot(rottp[idx, :])) + twopiaddition
            if localangles[idx] < (localangles[idx - 1] - np.pi):
                twopiaddition += 2 * np.pi
                localangles[idx] += 2 * np.pi

        self.averagescanangles = localangles
        self.xaxis = xaxis
        self.yaxis = yaxis

        totalanglechange = np.max(localangles)
        averageanglechange = totalanglechange / np.size(self.averagescanmjd) * self.numinbetween
        possibleretangles = np.linspace(0, totalanglechange, np.int(np.floor(totalanglechange / averageanglechange)))

        if self.placementtype == 'scan':
            possibleretangles = localangles
        elif self.placementtype == 'uniform':
            possibleretangles = possibleretangles[
                (possibleretangles > self.cutoff) & (possibleretangles < (totalanglechange - self.cutoff))]
        else:
            raise ValueError("Placement type is not one of ['scan', 'uniform']")

        self.retprofileloc = np.zeros((np.size(possibleretangles), 3))
        self.retprofileavglook = np.zeros((np.size(possibleretangles), 3))
        self.retprofilemjd = np.zeros(np.size(possibleretangles))
        self.numretprof = np.size(possibleretangles)
        self.angles = possibleretangles

        for idx in range(self.numretprof):
            self.retprofileloc[idx, :] = rotation_matrix(normal, -1 * possibleretangles[idx]).dot(rottp[0, :])

        self.retprofilemjd = np.interp(possibleretangles, localangles, self.averagescanmjd)
        f = interpolate.interp1d(localangles, self.averagescanlook, axis=0)
        self.retprofileavglook = f(possibleretangles)

        self.retprofileloc = correct_for_earths_rotation(self.retprofileloc, -1 * self.retprofilemjd,
                                                         -1 * self.averagescanmjd[0])

    def get_local_plane(self, loc: np.ndarray) -> (np.ndarray, np.ndarray, np.ndarray):
        """
        Returns back the local plane at a specific location.
        
        Parameters
        ----------
        loc : numpy array
            Shape (3,) tangent point 
        
        Returns
        -------
        angles : numpy array
            Angles in radians in the plane which correspond to the profile locations
            angles are measured with respect to loc
        angleidx : numpy array
            Indicies of the angles in the grid
        normalandreference : numpy array
            The normal and reference of the returned local plane, this is intended
            to be passed into ISKEngine with the 'normalandreference' option
            Result is shape(6,) with the first 3 elements being the normal
            and the last 3 being the reference
        """
        anglesbetween = np.zeros(self.numretprof)
        for idx in range(self.numretprof):
            anglesbetween[idx] = np.arccos(
                loc.dot(self.retprofileloc[idx, :]) / np.linalg.norm(loc) / np.linalg.norm(self.retprofileloc[idx, :]))

        minidx = np.argmin(anglesbetween)
        # Use loc as the reference (xaxis), and calculate the normal from the
        # tangent points near
        reference = loc / np.linalg.norm(loc)
        # TODO: THIS NEEDS TO BE FIXED
        # We need to get this vector from the tangent point locations, not the 
        # Satellite look direction
        avglook = -1 * self.retprofileavglook[minidx, :]
        normal = np.cross(avglook, reference)
        normal /= np.linalg.norm(normal)
        yaxis = np.cross(reference, normal)
        yaxis /= np.linalg.norm(yaxis)

        # Rotate the profile locations by the earth's rotation to find the
        # angles in the tangent plane
        # rottp = correct_for_earths_rotation( self.retprofileloc, self.retprofilemjd, self.retprofilemjd[minidx])
        # NEW METHOD: Project the retrieval profile locations onto the plane
        rottp = project_onto_plane(self.retprofileloc, normal)
        allangles = np.zeros(self.numretprof)
        twopiaddition = 0
        for idx in range(self.numretprof):
            allangles[idx] = np.arctan2(yaxis.dot(rottp[idx, :]), reference.dot(rottp[idx, :]))
            if idx > 0:
                if allangles[idx] < (allangles[idx - 1] - np.pi):
                    twopiaddition += 2 * np.pi
                    allangles[idx] += 2 * np.pi

        # Have to be careful because we do not assume periodicity at the edges
        # TODO: make this user set
        maxanglediff = np.deg2rad(20)

        lowangleidx = ((allangles >= -1 * maxanglediff) & (allangles < 0)).nonzero()[0]
        if np.size(lowangleidx) == 0:
            lowangleidx = 0
        else:
            lowangleidx = lowangleidx[0]
        highangleidx = (allangles[minidx:] > maxanglediff).nonzero()[0]
        if np.size(highangleidx) == 0:
            highangleidx = self.numretprof
        else:
            highangleidx = highangleidx[0] + minidx

        angleidx = np.arange(lowangleidx, highangleidx)
        if np.size(angleidx) == 0:
            angleidx = minidx

        localangles = allangles[angleidx]
        sortidx = np.argsort(localangles)
        localangles = localangles[sortidx]
        angleidx = angleidx[sortidx]

        # Return back what we need
        return localangles, angleidx, np.concatenate((normal, reference))

    def interpolate_onto_grid(self, prof_lon: np.ndarray, prof_lat: np.ndarray, prof_height: np.ndarray,
                              prof: np.ndarray, do_height_interp=True) -> np.ndarray:
        """
        Converts a lat/lon prof that is defined on any grid to the retrieval grid
        Parameters
        ----------
        prof_lon:
            Longitude for the profile
        prof_lat:
            Latitude for the profile
        prof_height:
            Heights for the profile.  If the profile is only two dimensions this is not used
        prof:
            The profile. (lat, lon, height)
        do_height_interp:
            True if you want to do height interpolation, false otherwise.

        Returns
        -------
        (lat, lon, height) profile on the retrieval grid
        """
        retlat, retlon, retheight = geometry.lat_lon_alts(self.retprofileloc)
        retlon[retlon > 180] -= 360
        prof_lon = np.asarray(prof_lon)
        prof_lon[prof_lon > 180] -= 360
        retheight = self.altitudes

        # First get vertical profiles at the retrieval locations
        s = np.shape(prof)

        if len(s) > 1:
            # The profile is a 2d profile
            temp_prof = np.zeros((len(retlat), s[1]))
            # profile is (location,height), so loop over heights
            for idx in range(s[1]):
                temp_prof[:, idx] = self.distance_interpolate(prof_lat, prof_lon, retlat,
                                                              retlon, prof[:, idx])

            if do_height_interp:
                # Now construct the retrieval profile by interpolating over height
                ret_prof = np.zeros((len(retlat), len(retheight)))
                for idx in range(len(retlat)):
                    ret_prof[idx, :] = np.interp(retheight, prof_height, temp_prof[idx, :])

                return ret_prof
            else:
                return temp_prof

        else:
            temp_prof = self.distance_interpolate(prof_lat, prof_lon, retlat, retlon, prof)
            return temp_prof

    def interpolate_from_grid(self, prof_lon: np.ndarray, prof_lat: np.ndarray, prof_height: np.ndarray,
                              ret_prof: np.ndarray, do_height_interp=True) -> np.ndarray:
        """
        Converts a profile that is specified on the retrieval grid to a new lat/lon grid
        Parameters
        ----------
        prof_lon:
            Longitude to interpolate onto
        prof_lat:
            Latitude to interpolate onto
        prof_height:
            Heights to interpolate onto
        ret_prof:
            Profile that is specified on the retrieval grid
        do_height_interp:
            True if height interpolation should be done

        Returns
        -------
        Profile on the new grid
        """
        retlat, retlon, retheight = geometry.lat_lon_alts(self.retprofileloc)
        retlon[retlon > 180] -= 360
        prof_lon = np.asarray(prof_lon)
        prof_lon[prof_lon > 180] -= 360
        retheight = self.altitudes

        # First get vertical profiles at the retrieval locations
        s = np.shape(ret_prof)

        if len(s) > 1:
            # The profile is a 2d profile
            temp_prof = np.zeros((len(retlat), s[1]))
            # profile is (location,height), so loop over heights
            for idx in range(s[1]):
                temp_prof[:, idx] = self.distance_interpolate(retlat, retlon, prof_lat,
                                                              prof_lon, ret_prof[:, idx])
            if do_height_interp:
                # Now construct the retrieval profile by interpolating over height
                meas_prof = np.zeros((len(prof_lat), len(prof_height)))
                for idx in range(len(prof_lat)):
                    meas_prof[idx, :] = np.interp(prof_height, retheight, temp_prof[idx, :])

                return meas_prof
            else:
                return temp_prof

        else:
            temp_prof = self.distance_interpolate(retlat, retlon, prof_lat, prof_lon, ret_prof)
            return temp_prof

    @staticmethod
    def distance_interpolate(from_lat: np.ndarray, from_lon: np.ndarray, to_lat: np.ndarray, to_lon: np.ndarray,
                             profile: np.ndarray) -> np.ndarray:
        new_profile = np.zeros((len(to_lat)))

        end_lon = from_lon[-1]
        end_lat = from_lat[-1]
        dist_grid_to_end = np.cos(np.deg2rad(end_lat)) * np.cos(np.deg2rad(from_lat)) * np.cos(
            np.deg2rad(from_lon - end_lon)) + np.sin(np.deg2rad(from_lat)) * np.sin(np.deg2rad(end_lat))
        dist_grid_to_end[dist_grid_to_end > 1] = 1
        dist_grid_to_end[dist_grid_to_end < -1] = -1
        dist_grid_to_end = np.arccos(dist_grid_to_end)

        for idx, (lat, lon) in enumerate(zip(to_lat, to_lon)):
            # Calculate the distance from the this point to all old locations
            # assuming a spherical earth
            distance = np.arccos(np.cos(np.deg2rad(lat)) * np.cos(np.deg2rad(from_lat)) * np.cos(
                np.deg2rad(from_lon - lon)) + np.sin(np.deg2rad(from_lat)) * np.sin(np.deg2rad(lat)))

            # Check if we are outside the grid
            if np.all(np.diff(distance) > 0):
                new_profile[idx] = profile[0]
                continue
            elif np.all(np.diff(distance) < 0):
                new_profile[idx] = profile[-1]
                continue

            # We have the distances of the grid points to our desired interp point, but we need to assign a sign
            # do to the actual interpolation
            dist_point_to_end = np.arccos(np.cos(np.deg2rad(lat)) * np.cos(np.deg2rad(end_lat)) * np.cos(
                np.deg2rad(end_lon - lon)) + np.sin(np.deg2rad(end_lat)) * np.sin(np.deg2rad(lat)))

            distance *= np.sign(dist_point_to_end - dist_grid_to_end)

            if not (np.all(np.diff(distance) > 0) or np.all(np.diff(distance) < 0)):
                logging.warning("Distance between points is not monotonic, using the nearest point instead")
                new_profile[idx] = sp.interpolate.interp1d(distance, profile, kind='nearest')(0.0)

            if distance[1] < distance[0]:
                new_profile[idx] = np.interp([0.0], distance[::-1], profile[::-1])
            else:
                new_profile[idx] = np.interp([0.0], distance, profile)

        return new_profile
