import copy as copy

import numpy as np
import scipy as sp
import scipy.io as io
from scipy import signal

import retrieval.species as species
import radtran.geometry as geometry
from retrieval.method import tikhonov as tikh
import utility as utility
import yaml


def filter_run(x, n=3):
    return signal.resample(x, len(x), window=('general_gaussian', 1, len(x) / n))


class TomographicOzone(species.Species):
    def __init__(self, atmosphere, apriori_scaling=False, horiz_tikh_factor=0, vertical_tikh_factor=0,
                 config_file='ozone_default.yaml'):

        yaml_file = open(config_file, "r")
        config_data = yaml.load(yaml_file)

        self.triplet = []
        self.normalt = []
        self.triplet_alt_weighting = []

        for triplet in config_data['triplets']:
            wavels = np.concatenate((triplet['ozone_wavel'], triplet['reference_wavel']))

            weights = [-1]
            if triplet['ref_weights'] == "Auto":
                if len(triplet['reference_wavel']) == 1:
                    weights.append(1)
                elif len(triplet['reference_wavel']) == 2:
                    w = np.abs((triplet['reference_wavel'][0] - triplet['ozone_wavel'][0]) / (triplet['reference_wavel'][0] - triplet['reference_wavel'][1]))
                    weights.append(w)
                    weights.append(1-w)
                else:
                    raise ValueError("Ozone triplets not configures correctly in " + config_file)
            else:
                for w in triplet['ref_weights']:
                    weights.append(w)
            self.triplet.append([tuple(wavels), tuple(weights)])
            self.normalt.append([triplet['normalization_alt']])
            self.triplet_alt_weighting.append([triplet['alt_weight_grid'], triplet['alt_weight_values']])

        self.y_normalization_mode = config_data['y_normalization_mode']
        self.normalt_idx_range = 5

        self.wavel_nm = None
        self.ident = 'tomographic_ozone'
        self.atmosphere = atmosphere
        self.apriori_scaling = apriori_scaling

        # Only use perturbations that fall in this range
        self.pidx = np.arange(4, 59)
        # TODO: Generalize
        self.tanalts = np.linspace(500, 99500, 100)
        self.lowalt = 12000
        self.highalt = 59000
        self.maxupdatefactor = 2
        self.maxvalue = None
        self.iter = []
        self.profile = []

        self.horiz_tikh_factor = horiz_tikh_factor
        self.vertical_tikh_factor = vertical_tikh_factor
        self.tp = None
        self.tanalt = None
        self.scanlat = None
        self.scanlon = None
        self.retlat = None
        self.retlon = None
        self.alpha_low_alt = None

        self.profile = self.atmosphere.get_user_defined_twodim(self.ident)

        yaml_file.close()

    def get_required_wavelengths(self, inst):
        """
        Calculates the required wavelengths for the retrieval
        
        Parameters
        ----------
        inst : 
            Instrument, should have wavel_nm parameter
        
        Returns
        -------
        wavel : numpy array
            Wavelengths required for the retrieval
        """
        wavel_nm = []
        for triplet in self.triplet:
            # Find the wavelength in the measured instrument closest to the
            # triplet wavelength
            for w in triplet[0]:
                wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm) - w))
                wavel_nm.append(inst.wavel_nm[wave_idx])
        wavel_nm = np.unique(np.array(wavel_nm))
        return wavel_nm

    def configure_instrument_and_method(self, ret_inst, meas_inst, method):
        # If lowalt wasnt set by the user, repeat it so its an array
        if not isinstance(self.lowalt, np.ndarray):
            self.lowalt = np.repeat(self.lowalt, len(meas_inst.radtran.scanslices))

            # Get the wavelengths we need for our triplet
        # and store it as an np array
        self.wavel_nm = self.get_required_wavelengths(meas_inst)

        ret_inst.set_wavelengths(self.wavel_nm)

        self.tp, self.tanalt = geometry.tangent_points(ret_inst.observer, ret_inst.los)
        self.tanalt = np.array(self.tanalt)

        W = self.weighting_matrix(self.tanalt, self.measurement_vector(meas_inst)[0], meas_inst.radtran.scanslices)



        method.W = W
        method.user_defined_weight = True

        self.scanlat = np.zeros(len(ret_inst.radtran.scanslices))
        self.scanlon = np.zeros(len(ret_inst.radtran.scanslices))
        # Calculate the latitudes, longitudes of the scans in the retrieval range
        for idx, s in enumerate(ret_inst.radtran.scanslices):
            tpslice = slice(s[0], s[-1])

            mean_tp = np.mean(np.array(self.tp[tpslice]), axis=0)

            self.scanlat[idx], self.scanlon[idx], alt = geometry.lat_lon_alts([mean_tp])

        self.retlat, self.retlon, alt = geometry.lat_lon_alts(meas_inst.radtran.retrieval_grid.retprofileloc)

        self.alpha_low_alt = meas_inst.radtran.retrieval_grid.interpolate_onto_grid(self.scanlon, self.scanlat, [], self.lowalt)

        # Set the perturbation indicies based on the alpha_low_alts

        # min_allowed_alt_m = np.min(self.alpha_low_alt)

        if self.horiz_tikh_factor != 0:
            numangle = ret_inst.radtran.retrieval_grid.numretprof
            numalt = len(self.pidx)
            gamma = tikh.two_dim_horizontal_second_deriv(numangle, numalt, factor=self.horiz_tikh_factor)
            gamma += tikh.two_dim_vertical_second_deriv(numangle, numalt, factor=self.vertical_tikh_factor)
            method.gamma = gamma
        else:
            # Even though None is already default in the method class, we explicitly do this in case the method
            # is being reused for multiple species
            method.gamma = np.zeros(1)

        ret_inst.radtran.set_wf_alts(self.tanalts[self.pidx], np.ones(len(self.tanalts[self.pidx]))*1000)

    def weighting_matrix(self, tanalt, y_meas, scanslices):
        """
        Returns the MART weighting matrix
        
        Parameters
        ----------
        tanalt : numpy array
            Tanalt in meters, same length as the number of lines of sight

        y_meas: numpy array
            Measurement vector for the observations.

        scanslices: list
        
        Returns
        -------
        W : 2d Numpy Array
            Shape (triplet, los) weighting matrix
        """
        w = np.zeros((len(tanalt), len(self.triplet)))
        # interpolate onto tangent altitudes
        for i in range(0, len(tanalt)):
            for j in range(0, len(self.triplet)):
                w[i, j] = np.interp(tanalt[i], self.triplet_alt_weighting[j][0], self.triplet_alt_weighting[j][1])

        if self.y_normalization_mode == "by_measurement":
            y_meas[np.isnan(y_meas)] = 1
            y_meas[y_meas == 0] = 1
            return sp.sparse.diags(np.sqrt(w.transpose().flatten() / np.abs(y_meas)), 0)
        elif self.y_normalization_mode == "none":
            return sp.sparse.diags(np.sqrt(w.transpose().flatten()), 0)
        else:
            raise ValueError("Invalid value for y_normalization_mode: " + self.y_normalization_mode +
                             "-- Should be one of 'by_measurement', or 'none'")

    def measurement_vector(self, inst, do_error=False):
        rad = np.transpose(inst.rad)
        if do_error:
            rad_error = np.transpose(inst.raderror)
        # rad is now [wavel, los]
        y = np.zeros((len(self.triplet), np.shape(rad)[1]))
        y_error = np.zeros((len(self.triplet), np.shape(rad)[1]))
        # y is now [triplet, los]
        scanslices = inst.radtran.scanslices

        for idx, triplet in enumerate(self.triplet):
            for wavel, coeff in zip(triplet[0], triplet[1]):
                # find the closest wavelength and get the radiance
                wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm) - wavel))
                for idz, s in enumerate(scanslices):
                    radI = rad[wave_idx, s]
                    alt_idx = np.nonzero(self.tanalt > self.normalt[idx][0])[0][0]
                    radR = np.mean(radI[alt_idx:(alt_idx + self.normalt_idx_range)])
                    y[idx, s] += coeff * np.log(radI / radR)

                    if do_error:
                        # The squared error in the actual measurements used
                        radI_error = (rad_error[wave_idx, s] / radI) ** 2

                        # The squared error in the normalization altitude
                        radR_error = np.sum(radI_error[alt_idx:(alt_idx + self.normalt_idx_range)]) / self.normalt_idx_range

                        y_error[idx, s] += (coeff**2) * (radI_error + radR_error)

                    bad = np.asarray(self.tanalt)[s] > self.tanalt[s[alt_idx]]
                    y[idx, s[bad]] = np.nan  # remove points above tangent altitude

                    bad = np.asarray(self.tanalt)[s] < self.lowalt[idz]
                    y[idx, s[bad]] = np.nan  # remove points below low altitude

        y_error = np.sqrt(y_error).flatten()
        return y.flatten(), y_error

    def measurement_vector_jacobian(self, inst):
        wf = inst.wf
        rad = np.transpose(inst.rad)
        # wf is [wavel, los, pert]
        # Rad is [wavel, los]
        # Jacobian has to be [triplet*los,pert]

        # Only want to retrieve on perturbation altitudes in pidx
        numangle = np.int(np.round(np.size(wf[0], axis=1) / len(self.pidx)))
        pertidx = np.array([])
        for idx in range(numangle):
            pertidx = np.concatenate((pertidx, self.pidx + 100 * idx))
        pertidx = pertidx.astype(int)

        numlos = np.size(wf[0], 0)

        # Divide by the radiance to get wf as dI/d(log)
        # s = np.shape(wf)
        # for idx in range(s[0]):
        #     s2 = np.shape(wf[idx])
        #     for idy in range(s2[0]):
        #         wf[idx][idy, :] /= self.profile.flatten()
        prof_factor = self.profile.flatten()[pertidx]

        J = sp.sparse.lil_matrix((len(self.triplet) * numlos, len(pertidx)))
        for idx, triplet in enumerate(self.triplet):
            triplet_range = slice(idx * numlos, (idx + 1) * numlos)
            t = np.zeros((numlos, len(pertidx)))
            for wavel, coeff in zip(triplet[0], triplet[1]):
                # find the closest wavelength and get the radiance
                wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm) - wavel))
                radI = rad[wave_idx, :]
                wfI = wf[wave_idx]
                vals = coeff * wfI.toarray() / radI[:, np.newaxis] * prof_factor[np.newaxis, :]
                vals[np.isnan(vals)] = 0
                t += vals
            J[triplet_range, :] = t

        self.good_alts = pertidx
        return sp.sparse.csr_matrix(J)

    def interp_alphas(self, alphas):
        """
        Interpolates alphas onto the profile retrieval grid.  Above the highest
        retrieval altitude we set alpha to a constant value.  Below the lowest
        retrieval altitude we fix alpha=1 at the ground and interpolate down
        
        Parameters
        ----------
        alphas : 2d numpy array
            Shape (angles, heights)
        Returns
        -------
        alphas : numpy array
            Interpolated alphas
        """
        alpha_interp = np.ones(np.shape(self.profile))
        for idx in range(np.size(self.profile, axis=0)):
            good = (self.tanalts[self.pidx] > self.alpha_low_alt[idx]) & (self.tanalts[self.pidx] < self.highalt)
            if self.apriori_scaling:
                alpha_alts = np.concatenate(([0], self.tanalts[self.pidx][good]))
                alpha = np.concatenate(([1], alphas[idx, good]))

                alpha_interp[idx, :] = np.interp(self.tanalts, alpha_alts, alpha)
            else:
                alpha_interp[idx, :] = np.interp(self.tanalts, self.tanalts[self.pidx][good], alphas[idx, good], left=1)

        return alpha_interp

    def update_profile(self, delta, mode=None):
        delt = delta.reshape(-1, len(self.pidx))
        delt[delt > 10] = 10
        # alphas = (self.profile[:, self.pidx] + delt) / self.profile[:, self.pidx]
        alphas = np.exp(delt)
        alpha_interp = self.interp_alphas(alphas)
        alpha_interp[alpha_interp > self.maxupdatefactor] = self.maxupdatefactor
        alpha_interp[alpha_interp < 1 / self.maxupdatefactor] = 1 / self.maxupdatefactor
        self.profile *= alpha_interp
        if self.maxvalue is not None:
            self.profile[self.profile > self.maxvalue] = self.maxvalue

        self.iter.append(self.profile)

    def get_atmospheric_profile(self):
        return copy.copy(self.profile)

    def set_atmospheric_profile(self, profile):
        self.profile = copy.copy(profile)
        self.atmosphere.modify_user_defined_twodim(self.ident, self.profile)

    def get_x(self):
        numangle = np.size(self.profile, axis=0)
        pertidx = np.array([])
        for idx in range(numangle):
            pertidx = np.concatenate((pertidx, self.pidx + 100 * idx))
        pertidx = pertidx.astype(int)
        return np.log(self.profile.flatten()[pertidx])

    def propagate_covariance(self, covar_matrix):
        """
        The retrieval method calculates covariance on x, this function propagates the covariance matrix onto the
        reported profile grid and reports back the error on the profile
        """
        rel_error = np.sqrt(covar_matrix.diagonal().flatten()).reshape((-1, len(self.pidx)))

        rel_error_interp = np.ones(np.shape(self.profile))
        for idx in range(np.size(self.profile, axis=0)):
            good = (self.tanalts[self.pidx] > self.alpha_low_alt[idx]) & (self.tanalts[self.pidx] < self.highalt)
            rel_error_alts = self.tanalts[self.pidx][good]
            temp = rel_error[idx, good]

            rel_error_interp[idx, :] = np.interp(self.tanalts, rel_error_alts, temp)

        abs_error = rel_error_interp * self.profile

        return abs_error


    def output_data(self):
        return self.profile, self.iter


class TomographicAerosol(TomographicOzone):
    """
    Tomographic aerosol species, basically the same as tomographic ozone
    except we let the profile change by a larger factor every iteration
    """
    def __init__(self, atmosphere, apriori_scaling=True, horiz_tikh_factor=0, vertical_tikh_factor=0):
        # TODO: setup triplets from a file
        self.triplet = []
        # self.triplet.append(([745.67, 510.00], [1, -1]))
        self.triplet.append(([745.67], [1]))

        # We set normalt to something here but it is recacalculated later based on the minimum of the measurement vec
        self.normalt = []
        self.normalt.append(([35000.0, 10000]))

        self.wavel_nm = None
        self.ident = 'tomographic_aerosol'
        self.atmosphere = atmosphere

        # Only use perturbations that fall in this range
        self.pidx = np.arange(0, 46)
        # TODO: Generalize
        self.tanalts = np.linspace(500, 99500, 100)
        self.maxupdatefactor = 30
        self.maxvalue = 1000
        self.normalt_idx_range = 5
        self.iter = []
        self.lowalt = 14000
        self.highalt = 32000
        self.apriori_scaling = apriori_scaling
        self.horiz_tikh_factor = horiz_tikh_factor
        self.vertical_tikh_factor = vertical_tikh_factor
        self.error = 0.01
        self.scan_alt_norm_idx = None

        self.profile = self.atmosphere.get_user_defined_twodim(self.ident)

    def configure_instrument_and_method(self, ret_inst, meas_inst, method):
        # If lowalt wasnt set by the user, repeat it so its an array
        if not isinstance(self.lowalt, np.ndarray):
            self.lowalt = np.repeat(self.lowalt, len(meas_inst.radtran.scanslices))

            # Get the wavelengths we need for our triplet
        # and store it as an np array
        self.wavel_nm = self.get_required_wavelengths(meas_inst)
        ret_inst.radtran.set_wf_alts(self.tanalts[self.pidx], np.ones(len(self.tanalts[self.pidx]))*1000)

        ret_inst.set_wavelengths(self.wavel_nm)

        self.tp, self.tanalt = geometry.tangent_points(ret_inst.observer, ret_inst.los)
        self.tanalt = np.array(self.tanalt)

        if self.scan_alt_norm_idx is None:
            # First set the aerosol profile to 0 to calculate y_rayleigh
            old_x = copy.copy(super().get_atmospheric_profile())
            super().set_atmospheric_profile(np.zeros(np.shape(old_x)))
            ret_inst.calculate_radiance()
            y_ray = self.measurement_vector_not_normalized(ret_inst)[0]
            # Reset the old aerosol profile
            super().set_atmospheric_profile(old_x)

            y_meas = self.measurement_vector_not_normalized(meas_inst)[0]

            self.scan_alt_norm_idx = []
            for idx, s in enumerate(ret_inst.radtran.scanslices):
                y_ray_scan = y_ray[s]
                y_meas_scan = y_meas[s]

                # Only check for altitude normalizations greater than 25km
                good = (self.tanalt[s] > 27000) & (self.tanalt[s] < 45000)
                meas_diff = y_meas_scan - y_ray_scan
                highalt_idx = np.nanargmin(np.abs(y_meas_scan[good] - y_ray_scan[good])) + np.nonzero(good)[0][0]
                min_y = meas_diff[highalt_idx]
                norm_idx = []
                # Add all points above highalt_idx that are within min_y
                # Only go 3 points on either side maximum
                idx = 1
                norm_idx.append(highalt_idx)
                while True:
                    if idx > 3:
                        break
                    if np.abs(meas_diff[highalt_idx + idx] - min_y) < self.error:
                        norm_idx.append(highalt_idx + idx)
                        idx += 1
                    else:
                        break
                idx = 1
                while True:
                    if idx > 3:
                        break
                    if np.abs(meas_diff[highalt_idx - idx] - min_y) < self.error:
                        norm_idx.append(highalt_idx - idx)
                        idx += 1
                    else:
                        break
                norm_idx = np.sort(norm_idx)
                norm_idx = norm_idx[(self.tanalt[s][norm_idx] > 27000) & (self.tanalt[s][norm_idx] < 45000)]

                self.scan_alt_norm_idx.append(norm_idx)

        W = self.weighting_matrix(self.tanalt, self.measurement_vector(meas_inst)[0])



        method.W = W
        method.user_defined_weight = True

        self.scanlat = np.zeros(len(ret_inst.radtran.scanslices))
        self.scanlon = np.zeros(len(ret_inst.radtran.scanslices))
        # Calculate the latitudes, longitudes of the scans in the retrieval range
        for idx, s in enumerate(ret_inst.radtran.scanslices):
            tpslice = slice(s[0], s[-1])

            mean_tp = np.mean(np.array(self.tp[tpslice]), axis=0)

            self.scanlat[idx], self.scanlon[idx], alt = geometry.lat_lon_alts([mean_tp])

        self.retlat, self.retlon, alt = geometry.lat_lon_alts(meas_inst.radtran.retrieval_grid.retprofileloc)

        self.alpha_low_alt = meas_inst.radtran.retrieval_grid.interpolate_onto_grid(self.scanlon, self.scanlat, [], self.lowalt)

        if self.horiz_tikh_factor != 0:
            numangle = ret_inst.radtran.retrieval_grid.numretprof
            numalt = len(self.pidx)
            gamma = tikh.two_dim_horizontal_second_deriv(numangle, numalt, factor=self.horiz_tikh_factor)
            gamma += tikh.two_dim_vertical_second_deriv(numangle, numalt, factor=self.vertical_tikh_factor)
            method.gamma = gamma
        else:
            # Even though None is already default in the method class, we explicitly do this in case the method
            # is being reused for multiple species
            method.gamma = np.zeros(1)


    def measurement_vector(self, inst, do_error=False):
        rad = np.transpose(inst.rad)
        # rad is now [wavel, los]

        if do_error:
            rad_error = inst.raderror.T

        y = np.zeros((len(self.triplet), np.shape(rad)[1]))
        # y is now [triplet, los]
        y_error = np.zeros((len(self.triplet), np.shape(rad)[1]))
        scanslices = inst.radtran.scanslices

        for idx, triplet in enumerate(self.triplet):
            for wavel, coeff in zip(triplet[0], triplet[1]):
                # find the closest wavelength and get the radiance
                wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm) - wavel))
                for idz, s in enumerate(scanslices):
                    radI = rad[wave_idx, s]
                    alt_idx = self.scan_alt_norm_idx[idz]
                    radR = np.mean(radI[alt_idx])
                    y[idx, s] += coeff * np.log(radI / radR)

                    if do_error:
                        # The squared error in the actual measurements used
                        radI_error = (rad_error[wave_idx, s] / radI) ** 2

                        # The squared error in the normalization altitude
                        radR_error = np.sum(radI_error[alt_idx]) / len(alt_idx)

                        y_error[idx, s] += (coeff**2) * (radI_error + radR_error)


                    bad = np.asarray(self.tanalt)[s] > self.tanalt[s[alt_idx[0]]]
                    y[idx, s[bad]] = np.nan  # remove points above tangent altitude

                    bad = np.asarray(self.tanalt)[s] < self.lowalt[idz]
                    y[idx, s[bad]] = np.nan  # remove points below low altitude
        y_error = np.sqrt(y_error.flatten())
        return y.flatten(), y_error

    def measurement_vector_not_normalized(self, inst):
        rad = np.transpose(inst.rad)
        # rad is now [wavel, los]
        y = np.zeros((len(self.triplet), np.shape(rad)[1]))
        # y is now [triplet, los]
        scanslices = inst.radtran.scanslices

        for idx, triplet in enumerate(self.triplet):
            for wavel, coeff in zip(triplet[0], triplet[1]):
                # find the closest wavelength and get the radiance
                wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm) - wavel))
                for idz, s in enumerate(scanslices):
                    radI = rad[wave_idx, s]
                    y[idx, s] += coeff * np.log(radI)

        return y.flatten(), np.zeros(np.size(y))

    def weighting_matrix(self, tanalt, y_meas):
        self.hidx = (np.array(tanalt) > 0) & (np.array(tanalt) < 60000)
        self.hidx = self.hidx.astype(int)
        self.hidx = np.repeat(self.hidx, len(self.triplet))
        w = sp.sparse.diags(self.hidx, 0)
        return w

    def interp_alphas(self, alphas):
        """
        Interpolates alphas onto the profile retrieval grid.  Above/below the highest/lowest
        retrieval altitudes alphas are kept constant
        
        Parameters
        ----------
        alphas : 2d numpy array
            Shape (angles, heights)
        Returns
        -------
        alphas : numpy array
            Interpolated alphas
        """
        alpha_interp = np.ones(np.shape(self.profile))
        for idx in range(np.size(self.profile, axis=0)):
            low_norm_alt = self.tanalts[self.scan_alt_norm_idx[idx][0]]
            good = (self.tanalts[self.pidx] > self.alpha_low_alt[idx]) & (self.tanalts[self.pidx] < low_norm_alt)
            if self.apriori_scaling:
                alpha_interp[idx, :] = np.interp(self.tanalts, self.tanalts[self.pidx][good], alphas[idx, good])
            else:
                alpha_interp[idx, :] = np.interp(self.tanalts, self.tanalts[self.pidx][good], alphas[idx, good], left=1)

        return alpha_interp

    def propagate_covariance(self, covar_matrix):
        """
        The retrieval method calculates covariance on x, this function propagates the covariance matrix onto the
        reported profile grid and reports back the error on the profile
        """
        rel_error = np.sqrt(covar_matrix.diagonal().flatten()).reshape((-1, len(self.pidx)))

        rel_error_interp = np.ones(np.shape(self.profile))
        for idx in range(np.size(self.profile, axis=0)):
            low_norm_alt = self.tanalts[self.scan_alt_norm_idx[idx][0]]
            good = (self.tanalts[self.pidx] > self.alpha_low_alt[idx]) & (self.tanalts[self.pidx] < low_norm_alt)
            rel_error_alts = self.tanalts[self.pidx][good]
            temp = rel_error[idx, good]

            rel_error_interp[idx, :] = np.interp(self.tanalts, rel_error_alts, temp)

        abs_error = rel_error_interp * self.profile

        return abs_error